/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Assign
Version:   1.0
Status:    Stable
Date:      January 2004
ID:        ID_ASSIGN
Category:  Command
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2004.  All rights reserved.
Keywords:  assign, assignment, directory, file, location, divert, path, route

###DESCRIPTION###
<p>The Assign command class is provided for the purpose of adding, deleting
and altering assignments in the system.  Assignments provide a &quot;virtual
directory&quot; service that is used to manage the roots of the file
system.  From a user perspective they can be very helpful when used to
provide directory short-cuts.  They are also used by the system to define
standard directory names at the root level of the file system (see the
&quot;system:config/assignments.cfg&quot; file for a list).</p>

<p>Keep in mind that this class is only for use in command lines and as a
scripting command.  When programming with a traditional language, the
functions in the FileSystem module should be used.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/data/config.h>
#include <pandora/files/assign.h>
#include <pandora/modules/filesystem.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_ASSIGN, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2004.  All rights reserved.",
   "January 2004",
   "Assign"
};

static struct KernelBase *KernelBase;
static struct FileSystemBase *FileSystemBase;
static struct StringsBase *StringsBase;
static OBJECTPTR modFileSystem = NULL, modStrings = NULL;
static OBJECTPTR AssignClass = 0;

/*****************************************************************************
** Class definition.
*/

struct Assign {
   OBJECT_HEADER         /* Class header */
   OBJECTID OutputID;
   LONG     Save;        /* Set to TRUE to save the assignment permanently */
   LONG     Static;      /* Set to TRUE to make the object static */

   PRIVATE_FIELDS
   STRING Replace;
   STRING Delete;
   STRING Add;
};

static ERROR ASSIGN_Activate(struct Assign *, APTR);
static ERROR ASSIGN_ClosingTag(struct Assign *, APTR);
static ERROR ASSIGN_Free(struct Assign *, APTR);
static ERROR ASSIGN_Init(struct Assign *, APTR);

static ERROR SET_AssignAdd(struct Assign *, STRING);
static ERROR SET_AssignDelete(struct Assign *, STRING);
static ERROR SET_AssignReplace(struct Assign *, STRING);

static struct FieldArray AssignFields[] = {
   { "Output", 0, FDF_OBJECTID|FDF_RW, 0, NULL, NULL },
   { "Save",   0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   { "Static", 0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   /*** Virtual fields ***/
   { "Add",     0, FDF_STRING|FDF_W,  0, NULL, SET_AssignAdd },
   { "Delete",  0, FDF_STRING|FDF_W,  0, NULL, SET_AssignDelete },
   { "Replace", 0, FDF_STRING|FDF_W,  0, NULL, SET_AssignReplace },
   END_FIELD
};

static struct ActionArray AssignActions[] = {
   { AC_Activate,   ASSIGN_Activate },
   { AC_ClosingTag, ASSIGN_ClosingTag },
   { AC_Free,       ASSIGN_Free },
   { AC_Init,       ASSIGN_Init },
   { NULL, NULL }
};

static void Print(struct Assign *, STRING);

static BYTE AssignUsage[] = {
"USAGE: assign [save] [add|replace|delete]=assign;path\n\
\n\
System assignments can be managed with the Assign command.  You\n\
can use it to add, replace and delete assignments in the system.\n\
\n\
Here are some examples:\n\
\n\
  assign add=\"mystuff;user:documents/;athene:documents/\"\n\
  assign save replace=\"mystuff;sdk:manual/\"\n\
  assign delete=\"mystuff\"\n
\n\
Use the \"save\" option to permanently save the assignment so that\n\
it can survive the shutdown of your computer."
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&AssignClass, NULL,
      FID_BaseClassID|TLONG, ID_ASSIGN,
      FID_Version|TFLOAT,    VER_ASSIGN,
      FID_Name|TPTR,         "Assign",
      FID_Category|TLONG,    CCF_COMMAND,
      FID_Usage|TPTR,        AssignUsage,
      FID_Actions|TPTR,      AssignActions,
      FID_Fields|TPTR,       AssignFields,
      FID_Size|TLONG,        sizeof(struct Assign),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (modFileSystem) { Action(AC_Free, modFileSystem, NULL); modFileSystem = NULL; }
   if (modStrings)    { Action(AC_Free, modStrings, NULL); modStrings = NULL; }
   if (AssignClass)   { Action(AC_Free, AssignClass, NULL); AssignClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Assign: Activate
*/

static ERROR ASSIGN_Activate(struct Assign *Self, APTR Void)
{
   UBYTE buffer[1024];
   WORD i;

   if (Self->Replace) {
      DPrintF("Assign:","[Replace] %s", Self->Replace);

      StrCopy(Self->Replace, buffer, sizeof(buffer));
      StrTranslate(buffer, sizeof(buffer), NULL);

      for (i=0; (buffer[i]) AND (buffer[i] != ';'); i++);

      if (buffer[i] IS ';') {
         buffer[i] = 0;
         SetAssign(AST_NAME, buffer, AST_PATH, buffer + i + 1, AST_FLAGS, ASSIGN_REPLACE | (Self->Save ? ASSIGN_SAVE : 0), TAGEND);
      }
      else return(ERR_Failed);
   }

   if (Self->Delete) {
      DPrintF("Assign:","[Delete] %s", Self->Replace);
      DeleteAssign(Self->Delete);
   }

   if (Self->Add) {
      DPrintF("Assign:","[Add] %s", Self->Replace);

      StrCopy(Self->Add, buffer, sizeof(buffer));
      StrTranslate(buffer, sizeof(buffer), NULL);

      for (i=0; (buffer[i]) AND (buffer[i] != ';'); i++);

      if (buffer[i] IS ';') {
         buffer[i] = 0;
         SetAssign(AST_NAME, buffer, AST_PATH, buffer + i + 1, AST_FLAGS, (Self->Save ? ASSIGN_SAVE : 0), TAGEND);
         Print(Self, "The assignment has been added.");
         return(ERR_Okay);
      }
      else {
         Print(Self, "The assignment is not defined correctly.  The correct definition is \"assign;path\".");
         return(ERR_Failed);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Assign: ClosingTag
*/

static ERROR ASSIGN_ClosingTag(struct Assign *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      acActivate(Self);
      acFree(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Assign: Free
*/

static ERROR ASSIGN_Free(struct Assign *Self, APTR Void)
{
   if (Self->Replace) FreeMemory(Self->Replace);
   if (Self->Delete) FreeMemory(Self->Delete);
   if (Self->Add) FreeMemory(Self->Add);
   return(ERR_Okay);
}

/*****************************************************************************
** Assign: Init
*/

static ERROR ASSIGN_Init(struct Assign *Self, APTR Void)
{
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Add
Short:  Adds a list of paths to an assignment name.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>Setting the Add field with a list of paths will add those paths to the
specified assignment.  This will have the effect of preserving any existing
paths that are attributed to the assignment, if it already exists.  If the
assignment does not already exist, a new one will be created from scratch.
The string format to use when setting this field is:</p>

<blockquote>&quot;assign;path1/;path2/;path3/;...&quot;</blockquote>

<p>For example:</P>

<blockquote>&quot;templates;scripts:boot/wintel/templates/&quot;</blockquote>

###END###

*****************************************************************************/

static ERROR SET_AssignAdd(struct Assign *Self, STRING Value)
{
   UBYTE name[80];
   WORD i;

   if (Self->Static) {
      Self->Add = StrClone(Value);
      return(ERR_Okay);
   }

   for (i=0; (Value[i]) AND (Value[i] != ';') AND (i < sizeof(name)-1); i++) name[i] = Value[i];
   name[i] = 0;

   if (Value[i] IS ';') {
      SetAssign(AST_NAME, name, AST_PATH, Value + i + 1, AST_FLAGS, (Self->Save ? ASSIGN_SAVE : 0), TAGEND);
      Print(Self, "The assignment has been added.");
      return(ERR_Okay);
   }
   else {
      Print(Self, "The assignment is not defined correctly.  The correct definition is \"assign;path\".");
      return(ERR_Failed);
   }
}

/*****************************************************************************

###FIELD###
Name:   Delete
Short:  Deletes an assignment from the system.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>Assignments can be deleted by writing this field with the name of the
assignment that you want to delete.  If multiple paths are hooked into the
assignment, they will all be deleted.</p>

###END###

*****************************************************************************/

ERROR SET_AssignDelete(struct Assign *Self, STRING Name)
{
   if (Self->Static) {
      Self->Delete = StrClone(Name);
      return(ERR_Okay);
   }

   if (DeleteAssign(Name) IS ERR_Okay) {
      Print(Self, "The assignment has been deleted.");
      return(ERR_Okay);
   }
   else{
      Print(Self, "The assignment could not be deleted.");
      return(ERR_Failed);
   }
}

/*****************************************************************************

###FIELD###
Name:   Replace
Short:  Replaces the path of an existing assignment with a new path.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>This field is used to replace assignments with completely new path lists.
It can also can be used to add new assignments if the specified assignment
name does not exist.  The string format to use when setting this field
is:</p>

<blockquote>&quot;assign;path1/;path2/;path3/;...&quot;</blockquote>

<p>For example:</p>

<blockquote>&quot;templates;scripts:boot/templates/;scripts:boot/wintel/templates/&quot;</blockquote>

###END###

*****************************************************************************/

ERROR SET_AssignReplace(struct Assign *Self, STRING Name)
{
   LONG i;

   if (Self->Static) {
      Self->Replace = StrClone(Name);
      return(ERR_Okay);
   }

   for (i=0; (Name[i] != 0) AND (Name[i] != ';'); i++);
   if (Name[i] IS ';') {
      Name[i] = 0;
         SetAssign(AST_NAME, Name, AST_PATH, Name + i + 1, AST_FLAGS, ASSIGN_REPLACE | (Self->Save ? ASSIGN_SAVE : 0), TAGEND);
      Name[i] = ';';

      Print(Self, "The assignment has been replaced.");
      return(ERR_Okay);
   }
   else {
      Print(Self, "The assignment is not defined correctly.  The correct definition is \"assign;path\".");
      return(ERR_Failed);
   }
}

/****************************************************************************

###FIELD###
Name:   Output
Short:  Result messages will be sent to the object referred to in this field.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>If this field is set to a valid ObjectID, text messages will be sent to
the object when the Assign process executes.  This can be helpful for
notifying the user of the results of a comand's execution.</p>

<p>The object receiving the message must be capable of understanding text
sent via data channels.  In most cases it is recommended that a Text
object is used for this purpose.</p>

###END###

****************************************************************************/

/*****************************************************************************
** Internal: Print()
*/

static void Print(struct Assign *Self, STRING Buffer)
{
   struct acDataChannel channel;

   if (Self->OutputID) {
      channel.ObjectID = Self->Head.UniqueID;
      channel.DataType = DATA_TEXT;
      channel.Version  = 1;
      channel.Buffer   = Buffer;
      for (channel.Size=0; Buffer[channel.Size]; channel.Size++);
      channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &channel);
   }
}
