/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Font
Version:   2.0
ID:        ID_FONT
Include:   pandora/graphics/font.h
Category:  Graphics
Date:      April 2003
Author:    Rocklyte Systems, some portions based on the FreeType Project.
Copyright: Rocklyte Systems, 1998-2003. All rights reserved.
Short:     Draws text in different type faces and styles.
Keywords:  font, graphics,  typeset, draw, text, string

###DESCRIPTION###
<p>The Font class is provided for the purpose of rendering strings to Bitmap
graphics. It supports standard effects such as bold, italic and underlined text,
along with extra features such as adjustable spacing, word alignment and outlining.
Fixed-point bitmap fonts are supported through the Windows .fon file format
and TrueType font files are supported for scaled font rendering.</p>

<P>Fonts must be stored in the "fonts:" directory in order to be recognised and
either in the "fixed" or "truetype" sub-directories as appropriate.  The process of
font installation and file management is managed by the "SystemFonts" object, a member
of the FontServer class.  If you need to install, remove or retrieve font
information (such as a list of available faces), please refer to the
FontServer documentation for more detail.</p>

<p>The Font class includes full support for the unicode character set
through its support for UTF-8.  This gives you the added benefit of being
able to support international character sets with ease, but you must be
careful not to use character codes above #127 without being sure that they
follow UTF-8 guidelines.  You can find out more about UTF-8 at this
<a href="http://www.cl.cam.ac.uk/~mgk25/unicode.html">web page</a>.</p>

<p>Initialisation of a new font object can be as simple as declaring its
<a href="#Point">Point</a> size and <a href="#Face">Face</a> name.
It is necessary to bear in mind that fonts are resistent to alteration after
initialisation, so all style and graphical selections must be set early.  For
example, it is not possible to change styling from regular to bold format
without the request being rejected.  To support multiple styles of the same
font, you therefore need to create a font object for every style that you
wish to support.  Basic features such as colour, the font string and text
positioning are not affected by these restrictions.</p>

<p>To draw a font string to a Bitmap object, start by setting the
<a href="#Bitmap">Bitmap</a> and <a href="#String">String</a> fields.  The
<a href="#XCoord">XCoord</a> and <a href="#YCoord">YCoord</a> fields determine
string positioning and you can also use the <a href="#Align">Align</a> field
to align a string to the right or center of the drawable area.</p>

<p>Finally, here is some clarification on some of the terminology
used in the Font class and further documentation:</p>

<dir>
<li>'Point' determines the size of a font.  The value is relative only to other point sizes of the same font face, i.e. two faces at the same point size are not necessarily the same height.</li>
<li>'Height' represents the 'vertical bearing' or point of the font, expressed as a pixel value.  The height does not cover for any leading at the top of the font, or the gutter space used for the tails on characters like 'g' and 'y'.</li>
<li>'Gutter' is the amount of space that a character can descend below the base line.  Characters like 'g' and 'y' are examples of characters that utilise the gutter space.  The gutter is also sometimes known as the "external leading" of a character.</li>
<li>'LineSpacing' is the recommended pixel distance between each line that is printed with the font.</li>
</dir>

###END###

*****************************************************************************/

static ERROR ConvertFontGraphic(struct Font *, struct BitmapCache *, LONG, UBYTE **);
static ERROR GenerateFontOutline(struct Font *, struct BitmapCache *, UBYTE **);
static WORD CheckBitmapCache(struct Font *Self, LONG Style);
static WORD CacheBitmapFont(struct Font *Self, OBJECTPTR file, LONG offset, struct winfnt_header_fields *face);

/*****************************************************************************
** Font class definition.
*/

static ERROR GET_Bold(struct Font *, LONG *);
static ERROR GET_Width(struct Font *, LONG *);
static ERROR GET_Opacity(struct Font *, FLOAT *);
static ERROR GET_YOffset(struct Font *, LONG *);

static ERROR SET_Bold(struct Font *, LONG *);
static ERROR SET_Colour(struct Font *, STRING);
static ERROR SET_ColourRGB(struct Font *, struct RGB *);
static ERROR SET_Flags(struct Font *, LONG *);
static ERROR SET_String(struct Font *, STRING);
static ERROR SET_Face(struct Font *, STRING);
static ERROR SET_Location(struct Font *, STRING);
static ERROR SET_Outline(struct Font *, STRING);
static ERROR SET_OutlineRGB(struct Font *, struct RGB *);
static ERROR SET_Opacity(struct Font *, FLOAT *);
static ERROR SET_Style(struct Font *, STRING);
static ERROR SET_Tabs(struct Font *, WORD *);
static ERROR SET_Underline(struct Font *, STRING);
static ERROR SET_UnderlineRGB(struct Font *, struct RGB *);

static struct FieldDef BaseFlags[] = {
   { "BOLD",         FTF_BOLD       }, { "ITALICS",   FTF_ITALICS },
   { "ANTIALIAS",    FTF_ANTIALIAS  }, { "WORDWRAP",  FTF_WORDWRAP },
   { "UNDERLINE",    FTF_UNDERLINE  }, { "HEAVYLINE", FTF_HEAVYLINE },
   { "UPDATEXY",     FTF_UPDATEXY   }, { "OUTLINE",   FTF_OUTLINE },
   { "QUICKALIAS",   FTF_QUICKALIAS }, { "ITALIC",    FTF_ITALIC },
   { "PREFERSCALED", FTF_PREFERSCALED }, { "PREFERFIXED", FTF_PREFERFIXED },
   { "CHARCLIP",     FTF_CHARCLIP },
   /*** Synonyms ***/
   { "NEEDSCALED", FTF_PREFERSCALED }, { "NEEDFIXED", FTF_PREFERFIXED },
   { "CLIP",       FTF_CHARCLIP },
   { NULL, NULL }
};

static struct FieldDef AlignFlags[] = {
   { "RIGHT",      ALIGN_RIGHT      }, { "LEFT",     ALIGN_LEFT },
   { "BOTTOM",     ALIGN_BOTTOM     }, { "TOP",      ALIGN_TOP },
   { "HORIZONTAL", ALIGN_HORIZONTAL }, { "VERTICAL", ALIGN_VERTICAL },
   { "CENTER",     ALIGN_CENTER     }, { "MIDDLE",   ALIGN_MIDDLE },
   { NULL, NULL }
};

static struct FieldArray FontFields[] = {
   { "Bitmap",          0, FDF_OBJECT|FDF_RW,    ID_BITMAP,        NULL, NULL },
   { "OutlineRGB",      0, FDF_POINTER|FDF_W,    0,                NULL, SET_OutlineRGB },
   { "UnderlineRGB",    0, FDF_POINTER|FDF_W,    0,                NULL, SET_UnderlineRGB },
   { "ColourRGB",       0, FDF_POINTER|FDF_W,    0,                NULL, SET_ColourRGB },
   { "String",          0, FDF_STRING|FDF_RW,    0,                NULL, SET_String },
   { "Flags",           0, FDF_LONGFLAGS|FDF_RW, (LONG)BaseFlags,  NULL, SET_Flags },
   { "Point",           0, FDF_LONG|FDF_RI,      0,                NULL, NULL },
   { "Gutter",          0, FDF_LONG|FDF_RI,      0,                NULL, NULL },
   { "GlyphSpacing",    0, FDF_LONG|FDF_RW,      0,                NULL, NULL },
   { "LineSpacing",     0, FDF_LONG|FDF_RW,      0,                NULL, NULL },
   { "LineCount",       0, FDF_LONG|FDF_R,       0,                NULL, NULL },
   { "XCoord",          0, FDF_LONG|FDF_RW,      0,                NULL, NULL },
   { "YCoord",          0, FDF_LONG|FDF_RW,      0,                NULL, NULL },
   { "TabSize",         0, FDF_LONG|FDF_RW,      0,                NULL, NULL },
   { "TotalCharacters", 0, FDF_LONG|FDF_R,       0,                NULL, NULL },
   { "WrapWidth",       0, FDF_LONG|FDF_RW,      0,                NULL, NULL },
   { "FixedWidth",      0, FDF_LONG|FDF_RW,      0,                NULL, NULL },
   { "Location",        0, FDF_STRING|FDF_RW,    0,                NULL, SET_Location },
   { "Height",          0, FDF_LONG|FDF_R,       0,                NULL, NULL },
   { "Style",           0, FDF_STRING|FDF_RI,    0,                NULL, SET_Style },
   { "Leading",         0, FDF_LONG|FDF_R,       0,                NULL, NULL },
   { "MaxHeight",       0, FDF_LONG|FDF_R,       0,                NULL, NULL },
   { "Angle",           0, FDF_DOUBLE|FDF_RW,    0,                NULL, NULL },
   { "Face",            0, FDF_STRING|FDF_RI,    0,                NULL, SET_Face },
   { "Align",           0, FDF_LONGFLAGS|FDF_RW, (LONG)AlignFlags, NULL, NULL },
   { "AlignWidth",      0, FDF_LONG|FDF_RW,      0,                NULL, NULL },
   { "AlignHeight",     0, FDF_LONG|FDF_RW,      0,                NULL, NULL },
   { "Tabs",            0, FDF_POINTER|FDF_RW,   0,                NULL, SET_Tabs },
   { "TabWidth",        0, FDF_LONG|FDF_RW,      0,                NULL, NULL },
   /*** Virtual fields ***/
   { "Bold",         0, FDF_LONG|FDF_RW,   0,                GET_Bold,     SET_Bold },
   { "Colour",       0, FDF_STRING|FDF_W,  0,                NULL,         SET_Colour },
   { "Opacity",      0, FDF_FLOAT|FDF_RW,  0,                GET_Opacity,  SET_Opacity },
   { "Outline",      0, FDF_STRING|FDF_W,  0,                NULL,         SET_Outline },
   { "StrWidth",     0, FDF_LONG|FDF_R,    0,                GET_Width,    NULL }, /* OBSOLETE: Use Width */
   { "Translucency", 0, FDF_FLOAT|FDF_RW,  0,                GET_Opacity,  SET_Opacity },
   { "Underline",    0, FDF_STRING|FDF_W,  0,                NULL,         SET_Underline },
   { "Width",        0, FDF_LONG|FDF_R,    0,                GET_Width,    NULL },
   { "YOffset",      0, FDF_LONG|FDF_R,    0,                GET_YOffset,  NULL },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR FONT_AccessObject(struct Font *, APTR);
static ERROR FONT_Draw(struct Font *, APTR);
static ERROR FONT_Free(struct Font *, APTR);
static ERROR FONT_Init(struct Font *, APTR);
static ERROR FONT_NewObject(struct Font *, APTR);

static struct ActionArray FontActions[] = {
   { AC_AccessObject, FONT_AccessObject },
   { AC_Draw,         FONT_Draw },
   { AC_Free,         FONT_Free },
   { AC_Init,         FONT_Init },
   { AC_NewObject,    FONT_NewObject },
   { NULL, NULL }
};

/*****************************************************************************
** Methods
*/

static ERROR FONT_CalcStringWidth(struct Font *, struct mtCalcStringWidth *);
static ERROR FONT_ConvertFontCoords(struct Font *, struct mtConvertFontCoords *);

static struct FunctionField argsCalcStringWidth[]   = { { "String", ARG_STR }, { "Length", ARG_LONG }, { "Width", ARG_LONG|ARG_RESULT }, { NULL, NULL } };
static struct FunctionField argsConvertFontCoords[] = { { "String", ARG_STR }, { "XCoord", ARG_LONG }, { "YCoord", ARG_LONG }, { "Column", ARG_LONG|ARG_RESULT }, { "Row", ARG_LONG|ARG_RESULT }, { NULL, NULL } };

static struct MethodArray FontMethods[] = {
   { MT_CalcStringWidth,   FONT_CalcStringWidth,   "CalcStringWidth",   argsCalcStringWidth,  sizeof(struct mtCalcStringWidth) },
   { MT_ConvertFontCoords, FONT_ConvertFontCoords, "ConvertFontCoords", argsConvertFontCoords, sizeof(struct mtConvertFontCoords) },
   { NULL, NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Function: AddFontClass()
*/

static ERROR AddFontClass(void)
{
   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&FontClass, NULL,
      FID_BaseClassID|TLONG,       ID_FONT,
      FID_Version|TFLOAT,          VER_FONT,
      FID_Name|TSTRING,            "Font",
      FID_Category|TLONG,          CCF_GRAPHICS,
      FID_FileExtension|TSTRING,   "font;fnt;tty;fon",
      FID_FileDescription|TSTRING, "Font",
      FID_Actions|TPTR,            FontActions,
      FID_Methods|TPTR,            FontMethods,
      FID_Fields|TPTR,             FontFields,
      FID_Size|TLONG,              sizeof(struct Font),
      TAGEND));
}

/****************************************************************************/

static LONG ReadWordLE(OBJECTPTR File)
{
   struct mtReadLE read;
   read.Size = sizeof(WORD);
   read.Result = 0;
   Action(MT_ReadLE, File, &read);
   return(read.Result);
}

/****************************************************************************/

static LONG ReadLongLE(OBJECTPTR File)
{
   struct mtReadLE read;
   read.Size = sizeof(LONG);
   read.Result = 0;
   Action(MT_ReadLE, File, &read);
   return(read.Result);
}

/*****************************************************************************
** Font: AccessObject
*/

static ERROR FONT_AccessObject(struct Font *Self, APTR Void)
{
   Self->ColourRGB    = &Self->prvColourRGB;
   Self->UnderlineRGB = &Self->prvUnderlineRGB;
   Self->OutlineRGB   = &Self->prvOutlineRGB;
   Self->Location     = Self->prvLocation;
   Self->Style        = Self->prvStyle;
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  CalcStringWidth
Short: Returns the pixel width of any given string in relation to a font's settings.

###DESCRIPTION###
<p>This method calculates the pixel width of any string in relation to a
font's object definition.  The routine takes into account any line feeds that
might be specified in the String, so if the String contains 8 lines, then the
width of the longest line will be returned.</p>

<p>Word wrapping will not be taken into account from the font, even
if it has been enabled in the object.</p>

###INPUT###
STRING/String: The string to be calculated.
LONG/Length:   The number of characters (give consideration to UTF-8 formatting) to be used in calculating the string length, or -1 if you want the entire string to be used.
LONG/Width:    The calculated width will be stored in this parameter before returning.

###RESULT###
ERR_Okay:          The width was successfully calculated.
ERR_Args:          Invalid arguments were specified.
ERR_ObjectCorrupt: The font object structure is corrupt.

###END###

*****************************************************************************/

static ERROR FONT_CalcStringWidth(struct Font *Self, struct mtCalcStringWidth *Args)
{
   STRING str;
   LONG length, lastlength, strlen, glyph, prevglyph;
   ULONG unicode;

   if ((!Args) OR (!Args->String)) return(ObjectError(Self, MT_CalcStringWidth, ERR_Args));

   Args->Width = 0;

   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) return(ERR_NotInitialised);

   str = Args->String;
   if ((strlen = Args->Length) < 0) for (strlen=0; str[strlen]; strlen++);

   length     = 0;
   lastlength = 0;
   prevglyph  = 0;
   while ((*str) AND (strlen > 0)) {
      if (*str IS '\n') {
         if (lastlength < length) lastlength = length; /* Compare lengths */
         length  = 0; /* Reset */
         str++;
      }
      else if (*str IS '\t') {
         WORD tabwidth;
         tabwidth = (Self->prvChar[' '].Advance + Self->GlyphSpacing) * Self->TabSize;
         if (tabwidth) length = ROUNDUP(length, tabwidth);
         str++;
      }
      else {
         str += GetUTF8(Self, str, &unicode);

         if (Self->FixedWidth > 0) length += Self->FixedWidth + Self->GlyphSpacing;
         else if ((unicode < 256) AND (Self->prvChar[unicode].Advance)) {
            length += Self->prvChar[unicode].Advance + Self->GlyphSpacing;
         }
         else if (Self->Flags & FTF_SCALABLE) {
            if (unicode IS ' ') {
               glyph = prevglyph;

               if (Self->FixedWidth > 0) length += Self->FixedWidth + Self->GlyphSpacing;
               else length += Self->prvChar[' '].Advance + Self->GlyphSpacing;
            }
            else {
               glyph = FT_Get_Char_Index(Self->FTFace, unicode);

               if (!FT_Load_Glyph(Self->FTFace, glyph, FT_LOAD_DEFAULT)) {
                  length += (Self->FTFace->glyph->advance.x>>FT_DOWNSIZE) + Self->GlyphSpacing;
               }

               if (Self->Flags & FTF_KERNING) length += GetKerning(Self->FTFace, glyph, prevglyph);
            }

            prevglyph = glyph;
         }
         else length += Self->prvChar[Self->prvDefaultChar].Advance + Self->GlyphSpacing;
      }
      strlen--;
   }

   if (lastlength > length) Args->Width = lastlength - Self->GlyphSpacing;
   else if (length > 0) Args->Width = length - Self->GlyphSpacing;

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    ConvertFontCoords()
Short:   Converts pixel coordinates into equivalent column and row positions in font strings.

###DESCRIPTION###
<p>This method is used to convert pixel coordinates within a font String into
the equivalent Row and Column character positions.  If the coordinate values
that you supply are in excess of the String dimensions, the Column and Row
results will be automatically restricted to their maximum value.  For
instance, if the YCoord argument is set to 280 and the String consists of 15
rows amounting to 150 pixels in height, the Row value will be returned as 15.</p>

<p>Negative coordinate values are not permitted.</p>

###INPUT###
STRING/String: Must point to the string that you will be inspecting, or NULL if you want to inspect the string currently in the font's String field.
LONG/XCoord: The horizontal coordinate that you want to translate into a column position.
LONG/YCoord: The vertical coordinate that you want to translate into a row position.
LONG/Column: This parameter will be updated to reflect the calculated character position, with consideration to the UTF-8 standard.
LONG/Row: This parameter will be updated to reflect the calculated row position.
LONG/ByteColumn: This parameter will be updated to reflect the absolute column byte position within the given row.
LONG/BytePosition: This parameter will be updated to reflect the absolute byte position with the font String.

###RESULT###
ERR_Okay: The character position was calculated.
ERR_Args: Invalid arguments were specified.
ERR_FieldNotSet: The String field has not been set.

###END###

*****************************************************************************/

static ERROR FONT_ConvertFontCoords(struct Font *Self, struct mtConvertFontCoords *Args)
{
   STRING str;
   LONG y, i, width, xpos;
   ULONG unicode, prevglyph, glyph;

   if (!Args) return(ERR_Args);
   if ((Args->XCoord < 0) OR (Args->YCoord < 0)) return(ObjectError(Self, MT_ConvertFontCoords, ERR_Args));

   if (!(str = Args->String)) {
      if (!(str = Self->String)) return(ObjectError(Self, MT_ConvertFontCoords, ERR_FieldNotSet));
   }

   Args->Row    = 0;
   Args->Column = 0;
   Args->ByteColumn   = 0;
   Args->BytePosition = 0;

   /*** Calculate the row ***/

   y = Args->YCoord;
   while (y > Self->LineSpacing) {
      /*** Search for line feeds ***/
      i = 0;
      while ((str[i]) AND (str[i] != '\n')) for (++i; ((str[i] & 0xc0) IS 0x80); i++);
      if (str[i] IS '\n') {
         y -= Self->LineSpacing;
         Args->Row++;
         str += i + 1;
         Args->BytePosition += i + 1;
      }
      else break;
   }

   /*** Calculate the column ***/

   xpos = 0;
   width = 0;
   prevglyph = 0;
   while ((*str) AND (*str != '\n')) {
      if (Self->FixedWidth > 0) {
         str += GetUTF8(Self, str, NULL);
         width = Self->FixedWidth + Self->GlyphSpacing;
      }
      else if (*str IS '\t') {
         WORD tabwidth;
         tabwidth = (Self->prvChar[' '].Advance + Self->GlyphSpacing) * Self->TabSize;
         width = ROUNDUP(xpos, tabwidth) - xpos;
         str++;
      }
      else {
         str += GetUTF8(Self, str, &unicode);

         if ((unicode < 256) AND (Self->prvChar[unicode].Advance)) width = Self->prvChar[unicode].Advance + Self->GlyphSpacing;
         else if (Self->Flags & FTF_SCALABLE) {

            if (unicode IS ' ') {
               glyph = prevglyph;

               if (Self->FixedWidth > 0) width = Self->FixedWidth + Self->GlyphSpacing;
               else width = Self->prvChar[' '].Advance + Self->GlyphSpacing;
            }
            else {
               glyph = FT_Get_Char_Index(Self->FTFace, unicode);

               if (!FT_Load_Glyph(Self->FTFace, glyph, FT_LOAD_DEFAULT)) {
                  width = (Self->FTFace->glyph->advance.x>>FT_DOWNSIZE) + Self->GlyphSpacing;
               }

               if (Self->Flags & FTF_KERNING) width += GetKerning(Self->FTFace, glyph, prevglyph);
            }

            prevglyph = glyph;
         }
         else width = Self->prvChar[Self->prvDefaultChar].Advance + Self->GlyphSpacing;
      }

      /* Subtract the width of the current character and keep processing.
      ** Note that the purpose of dividing the width by 2 is to allow for
      ** rounding up if the point is closer to the right hand side of the
      ** character.
      */

      if (xpos + (width/2) >= Args->XCoord) break;
      xpos += width;

      Args->Column++;
      Args->ByteColumn++;
      Args->BytePosition++;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Draw
Short: Draws a font to a Bitmap.

###DESCRIPTION###
<p>When you are ready to draw a font to a Bitmap, use the Draw action.
Drawing will start from the coordinates given in the <a
href="#XCoord">XCoord</a> and <a href="#XCoord">YCoord</a> fields,
using the characters in the font <a href="#String">String</a>.  The
result of calling Draw will depend on what type of Font you are using and
how you have set up the font object.</p>

###RESULT###
ERR_Okay:        The font graphic was drawn successfully.
ERR_FieldNotSet: The Bitmap and/or String field has not been set.

###END###

*****************************************************************************/

static ERROR DrawBitmapFont(struct Font *);

static ERROR FONT_Draw(struct Font *Self, APTR Void)
{
   struct Bitmap *Bitmap;
   struct RGB rgb;
   UBYTE *str, *data, *tmp;
   LONG height, widestline, glyph, prevglyph, strwidth, offset;
   ULONG unicode;
   WORD startx, dxcoord, dycoord, locked, dx, dy, rendermode, ex, ey, sx, sy, xinc;
   FT_Matrix matrix;
   FT_Vector vector;

   if (!(Self->Flags & FTF_SCALABLE)) {
      /*** Draw unscaled bitmap font ***/
      return(DrawBitmapFont(Self));
   }

   /*** Validate settings for scaled font type ***/

   if (!(Bitmap = Self->Bitmap)) { DPrintF("@Draw:","[Font] The Bitmap field is not set."); return(ERR_FieldNotSet); }
   if (!Self->String) return(ERR_FieldNotSet);
   if (!Self->String[0]) return(ERR_Okay);

   if (!Self->LineCount) CalcLineCount(Self);

   rgb     = Self->prvColourRGB;
   str     = Self->String;
   dxcoord = Self->XCoord;
   dycoord = Self->YCoord;
   height  = (Self->MaxHeight + (Self->LineSpacing * (Self->LineCount-1))); //Self->LineCount * Self->LineSpacing;
   locked  = FALSE;

   if (!Self->AlignWidth)  Self->AlignWidth  = Bitmap->Width;
   if (!Self->AlignHeight) Self->AlignHeight = Bitmap->Height;

   if (Self->Angle) {
      DOUBLE radian;
      radian    = (Self->Angle * 3.141592654) / 180.0;
      matrix.xx = (FT_Fixed)(cos(radian) * 0x10000);
      matrix.xy = (FT_Fixed)(-sin(radian) * 0x10000);
      matrix.yx = (FT_Fixed)(sin(radian) * 0x10000);
      matrix.yy = (FT_Fixed)(cos(radian) * 0x10000);
      vector.x  = 0;
      vector.y  = 0;
   }

   GetField(Self, FID_Width, FT_LONG, &widestline);

   GET_YOffset(Self, &offset);
   dycoord += offset;

   /* If horizontal centering is required, calculate the correct
   ** horizontal starting coordinate.
   */

   if (Self->AlignFlags & (ALIGN_HORIZONTAL|ALIGN_RIGHT)) {
      tmp       = str;
      prevglyph = 0;
      glyph     = 0;
      strwidth  = 0;
      while ((*tmp) AND (*tmp != '\n')) {
         tmp += GetUTF8(Self, tmp, &unicode);

         if (Self->FixedWidth > 0) strwidth += Self->FixedWidth + Self->GlyphSpacing;
         else {
            if (unicode IS ' ') strwidth += Self->prvChar[' '].Advance;
            else {
               glyph = FT_Get_Char_Index(Self->FTFace, unicode);

               if (!FT_Load_Glyph(Self->FTFace, glyph, FT_LOAD_DEFAULT)) {
                  if (Self->Flags & FTF_KERNING) strwidth += GetKerning(Self->FTFace, glyph, prevglyph);
                  strwidth += (Self->FTFace->glyph->advance.x>>FT_DOWNSIZE) + Self->GlyphSpacing;
               }
            }

            prevglyph = glyph;
         }
      }

      if ((strwidth) AND (!Self->Angle)) {
         if (Self->AlignFlags & ALIGN_HORIZONTAL) dxcoord = Self->XCoord + ((Self->AlignWidth - strwidth) / 2);
         else dxcoord = Self->XCoord + Self->AlignWidth - strwidth;
      }
   }

   /*** Return immediately if the font is outside of the clipping region ***/

   #undef PIXEL_MARGIN
   #define PIXEL_MARGIN 10 /* The added pixel margin caters for gutter space, internal leading, outlining etc */

   if (dycoord - PIXEL_MARGIN >= Bitmap->ClipBottom) goto exit;
   if (dycoord + height + PIXEL_MARGIN <= Bitmap->ClipTop) goto exit;
   if (dxcoord - PIXEL_MARGIN>= Bitmap->ClipRight) goto exit;
   if (dxcoord + widestline + PIXEL_MARGIN <= Bitmap->ClipLeft) goto exit;

   /*** Grab the bitmap for direct pixel access ***/

   if (Action(AC_Lock, Bitmap, NULL) IS ERR_Okay) locked = TRUE;
   else {
      ObjectError(Self, ERH_Draw, ERR_LockFailed);
      goto exit;
   }

   if (Self->Flags & FTF_ANTIALIAS) rendermode = ft_render_mode_normal;
   else if (Self->Flags & FTF_QUICKALIAS) rendermode = ft_render_mode_normal;
   else rendermode = ft_render_mode_mono;

   if (Self->Flags & FTF_DIRECTXY) dycoord += Self->Leading;

   prevglyph = 0;
   glyph = 0;
   startx = dxcoord;
   while (*str) {
      if (*str IS '\n') {
         /*** Reset the font to a new line ***/

         if (Self->Flags & (FTF_UNDERLINE|FTF_HEAVYLINE)) {
            ActionTags(MT_DrawRGBLine, Bitmap, startx, dycoord + Self->Height, dxcoord, dycoord + Self->Height,
               Self->prvUnderlineRGB.Red, Self->prvUnderlineRGB.Green, Self->prvUnderlineRGB.Blue);
         }

         if (Self->Flags & FTF_HEAVYLINE) {
            ActionTags(MT_DrawRGBLine, Bitmap, startx, dycoord + Self->Height + 1, dxcoord, dycoord + Self->Height + 1,
               Self->prvUnderlineRGB.Red, Self->prvUnderlineRGB.Green, Self->prvUnderlineRGB.Blue);
         }

         str++;

         if (Self->AlignFlags & (ALIGN_HORIZONTAL|ALIGN_RIGHT)) {
            tmp       = str;
            prevglyph = 0;
            glyph     = 0;
            strwidth  = 0;
            while ((*tmp) AND (*tmp != '\n')) {
               tmp += GetUTF8(Self, tmp, &unicode);

               if (Self->FixedWidth > 0) strwidth += Self->FixedWidth + Self->GlyphSpacing;
               else {
                  if (unicode IS ' ') strwidth += Self->prvChar[' '].Advance;
                  else {
                     glyph = FT_Get_Char_Index(Self->FTFace, unicode);

                     if (!FT_Load_Glyph(Self->FTFace, glyph, FT_LOAD_DEFAULT)) {
                        if (Self->Flags & FTF_KERNING) strwidth += GetKerning(Self->FTFace, glyph, prevglyph);
                        strwidth += (Self->FTFace->glyph->advance.x>>FT_DOWNSIZE) + Self->GlyphSpacing;
                     }
                  }
                  prevglyph = glyph;
               }
            }

            if (strwidth) {
               if (Self->AlignFlags & ALIGN_HORIZONTAL) dxcoord = Self->XCoord + ((Self->AlignWidth - strwidth) / 2);
               else dxcoord = Self->XCoord + Self->AlignWidth - strwidth;
            }
            else dxcoord = Self->XCoord;
         }
         else dxcoord = Self->XCoord;

         startx = dxcoord;
         dycoord += Self->LineSpacing;
         prevglyph = 0;

         if (Self->Angle) {
            vector.x = dxcoord<<FT_DOWNSIZE;
            vector.y = dycoord<<FT_DOWNSIZE;
         }
      }
      else if (*str IS '\t') {
         WORD tabwidth;
         tabwidth = (Self->prvChar[' '].Advance + Self->GlyphSpacing) * Self->TabSize;
         dxcoord = Self->XCoord + ROUNDUP(dxcoord - Self->XCoord, tabwidth);
         str++;
         prevglyph = 0;
      }
      else {
         str += GetUTF8(Self, str, &unicode);

         if (Self->Angle) {
            FT_Set_Transform(Self->FTFace, &matrix, &vector);
         }

         if (unicode IS ' ') {
            glyph = prevglyph;
            if (Self->Angle) {
               vector.x += Self->FTFace->glyph->advance.x + (Self->GlyphSpacing<<FT_DOWNSIZE);
               vector.y += Self->FTFace->glyph->advance.y + (Self->GlyphSpacing<<FT_DOWNSIZE);
            }
            else {
               if (Self->FixedWidth > 0) dxcoord += Self->FixedWidth + Self->GlyphSpacing;
               else dxcoord += Self->prvChar[' '].Advance + Self->GlyphSpacing;
            }
         }
         else {
            glyph = FT_Get_Char_Index(Self->FTFace, unicode);
            if (FT_Load_Glyph(Self->FTFace, glyph, FT_LOAD_DEFAULT)) break;
            if (FT_Render_Glyph(Self->FTFace->glyph, rendermode)) break;

            if (Self->Flags & FTF_KERNING) {
               if (Self->Angle) {
               }
               else dxcoord += GetKerning(Self->FTFace, glyph, prevglyph);
            }

            if (Self->FTFace->glyph->bitmap.pixel_mode != FT_PIXEL_MODE_GRAY) break;

            data = Self->FTFace->glyph->bitmap.buffer;

            sx = dxcoord + Self->FTFace->glyph->bitmap_left;
            //if (Self->Angle) sx += dxcoord;
            ex = sx + Self->FTFace->glyph->bitmap.width;

            if (ex > Bitmap->ClipRight) ex = Bitmap->ClipRight;

            if (sx < Bitmap->ClipLeft) {
               data += Bitmap->ClipLeft - sx;
               sx = Bitmap->ClipLeft;
            }

            sy = dycoord - Self->FTFace->glyph->bitmap_top + Self->Height;

            //if (Self->Angle) sy += dycoord;
            ey = sy + Self->FTFace->glyph->bitmap.rows;

            if (ey > Bitmap->ClipBottom) ey = Bitmap->ClipBottom;

            if (sy < Bitmap->ClipTop) {
               data += Self->FTFace->glyph->bitmap.pitch * (Bitmap->ClipTop - sy);
               sy = Bitmap->ClipTop;
            }

            sx += Bitmap->XOffset; /* Add offsets only after clipping adjustments */
            sy += Bitmap->YOffset;
            dx += Bitmap->XOffset;
            dy += Bitmap->YOffset;
            ex += Bitmap->XOffset;
            ey += Bitmap->YOffset;

            xinc = (Self->FTFace->glyph->bitmap.pitch - Self->FTFace->glyph->bitmap.width) + (Self->FTFace->glyph->bitmap.width - (ex - sx));

            if (Self->Flags & FTF_QUICKALIAS) {
               for (dy=sy; dy < ey; dy++) {
                  for (dx=sx; dx < ex; dx++) {
                     if (data[0] > 2) {
                        rgb.Red   = (Self->prvColourRGB.Red   * data[0])>>8;
                        rgb.Green = (Self->prvColourRGB.Green * data[0])>>8;
                        rgb.Blue  = (Self->prvColourRGB.Blue  * data[0])>>8;
                        Bitmap->DrawUCRPixel(Bitmap, dx, dy, &rgb);
                     }
                     data++;
                  }
                  data += xinc;
               }
            }
            else {
               struct BRGB brgb;
               UBYTE *srctable, *desttable, trans, *bitdata, *line;

               line = Bitmap->Data + (sy * Bitmap->LineWidth) + (sx * Bitmap->BytesPerPixel);
               for (dy=sy; dy < ey; dy++) {
                  bitdata = line;
                  for (dx=sx; dx < ex; dx++) {
                     if (data[0] > 2) {
                        trans = (glAlphaLookup + (data[0]<<8))[Self->prvTranslucency]; /* Multiply the source alpha level by overall translucency level */
                        srctable  = glAlphaLookup + (trans<<8);
                        desttable = glAlphaLookup + ((255-trans)<<8);

                        Bitmap->ReadUCRIndex(Bitmap, bitdata, &brgb);

                        brgb.Red   = srctable[Self->prvColourRGB.Red]   + desttable[brgb.Red];
                        brgb.Green = srctable[Self->prvColourRGB.Green] + desttable[brgb.Green];
                        brgb.Blue  = srctable[Self->prvColourRGB.Blue]  + desttable[brgb.Blue];

                        Bitmap->DrawUCRIndex(Bitmap, bitdata, &brgb);
                     }
                     bitdata += Bitmap->BytesPerPixel;
                     data++;
                  }
                  line += Bitmap->LineWidth;
                  data += xinc;
               }
            }

            if (Self->Angle) {
               vector.x += Self->FTFace->glyph->advance.x + (Self->GlyphSpacing<<FT_DOWNSIZE);
               vector.y += Self->FTFace->glyph->advance.y + (Self->GlyphSpacing<<FT_DOWNSIZE);
               //dxcoord = vector.x>>FT_DOWNSIZE;
               //dycoord = vector.y>>FT_DOWNSIZE;
            }
            else {
               if (Self->FixedWidth > 0) dxcoord += Self->FixedWidth + Self->GlyphSpacing;
               else dxcoord += (Self->FTFace->glyph->advance.x>>FT_DOWNSIZE) + Self->GlyphSpacing;
            }
         }

         prevglyph = glyph;
      }
   }

   /*** Draw an underline for the current line if underlining is turned on ***/

   if (Self->Flags & (FTF_UNDERLINE|FTF_HEAVYLINE)) {
      ActionTags(MT_DrawRGBLine, Bitmap, startx, dycoord + Self->Height, dxcoord, dycoord + Self->Height,
         Self->prvUnderlineRGB.Red, Self->prvUnderlineRGB.Green, Self->prvUnderlineRGB.Blue);
   }

   if (Self->Flags & FTF_HEAVYLINE) {
      ActionTags(MT_DrawRGBLine, Bitmap, startx, dycoord + Self->Height + 1, dxcoord, dycoord + Self->Height + 1,
         Self->prvUnderlineRGB.Red, Self->prvUnderlineRGB.Green, Self->prvUnderlineRGB.Blue);
   }

   /*** Update X/Y coordinates if requested ***/

   if (Self->Flags & FTF_UPDATEXY) {
      Self->XCoord = dxcoord;
      Self->YCoord = dycoord;
   }

exit:
   if (locked IS TRUE) Action(AC_Unlock, Bitmap, NULL);

   return(ERR_Okay);
}

/****************************************************************************/

static ERROR DrawBitmapFont(struct Font *Self)
{
   struct mtGetColour getcolour;
   struct Bitmap *bitmap;
   struct RGB rgb, outlinergb;
   UBYTE *str, *data, *tmp;
   LONG height, widestline, strwidth, offset;
   ULONG unicode;
   WORD startx, dxcoord, dycoord, xpos, dx, dy, ex, ey, sx, sy, xinc, bytewidth, alpha;
   BYTE charclip_count;

   /*** Validate settings for fixed font type ***/

   if (!(bitmap = Self->Bitmap)) { DPrintF("@Draw:","[Font] The Bitmap field is not set."); return(ERR_FieldNotSet); }
   if (!Self->String) return(ERR_FieldNotSet);
   if (!Self->String[0]) return(ERR_Okay);

   if (!Self->LineCount) CalcLineCount(Self);

   outlinergb = Self->prvOutlineRGB;
   rgb     = Self->prvColourRGB;
   str     = Self->String;
   dxcoord = Self->XCoord;
   dycoord = Self->YCoord;
   height  = (Self->MaxHeight + (Self->LineSpacing * (Self->LineCount-1))); //Self->LineCount * Self->LineSpacing;
   charclip_count = 0;

   if (!Self->AlignWidth)  Self->AlignWidth  = bitmap->Width;
   if (!Self->AlignHeight) Self->AlignHeight = bitmap->Height;

   GET_YOffset(Self, &offset);
   dycoord += offset;

   /* If horizontal centering is required, calculate the correct
   ** horizontal starting coordinate.
   */

   GetField(Self, FID_Width, FT_LONG, &widestline);

   if (Self->AlignFlags & (ALIGN_HORIZONTAL|ALIGN_RIGHT)) {
      tmp      = str;
      strwidth  = 0;
      while ((*tmp) AND (*tmp != '\n')) {
         tmp += GetUTF8(Self, tmp, &unicode);

         if (Self->FixedWidth > 0) strwidth += Self->FixedWidth + Self->GlyphSpacing;
         else {
            if (unicode < 256) {
               strwidth += Self->prvChar[unicode].Advance + Self->GlyphSpacing;
            }
         }
      }

      if (strwidth) {
         if (Self->AlignFlags & ALIGN_HORIZONTAL) dxcoord = Self->XCoord + ((Self->AlignWidth - strwidth) / 2);
         else dxcoord = Self->XCoord + Self->AlignWidth - strwidth;
      }
   }

   /*** Return immediately if the font is outside of the clipping region ***/

   #undef PIXEL_MARGIN
   #define PIXEL_MARGIN 1 /* The added pixel margin caters for gutter space, internal leading, outlining etc */

   if ((Self->LineCount <= 1) AND (!(Self->AlignFlags & (ALIGN_HORIZONTAL|ALIGN_RIGHT)))) {
      if (dxcoord - PIXEL_MARGIN >= bitmap->ClipRight) return(ERR_Okay);
      if (dxcoord + widestline + PIXEL_MARGIN <= bitmap->ClipLeft)  return(ERR_Okay);
   }

   if (dycoord - PIXEL_MARGIN >= bitmap->ClipBottom) return(ERR_Okay);
   if (dycoord + height + PIXEL_MARGIN <= bitmap->ClipTop) return(ERR_Okay);

   getcolour.Red   = Self->prvColourRGB.Red;
   getcolour.Green = Self->prvColourRGB.Green;
   getcolour.Blue  = Self->prvColourRGB.Blue;
   getcolour.Alpha = 0;
   Action(MT_GetColour, bitmap, &getcolour);

   if (Action(AC_Lock, bitmap, NULL) IS ERR_Okay) {
      startx = dxcoord;
      alpha = 255 - Self->prvTranslucency;
      while (*str) {
         if (*str IS '\n') {
            /*** Reset the font to a new line ***/

            if (Self->Flags & (FTF_UNDERLINE|FTF_HEAVYLINE)) {
               ActionTags(MT_DrawRGBLine, bitmap, startx, dycoord + Self->Height, dxcoord, dycoord + Self->Height,
                  Self->prvUnderlineRGB.Red, Self->prvUnderlineRGB.Green, Self->prvUnderlineRGB.Blue);
            }

            if (Self->Flags & FTF_HEAVYLINE) {
               ActionTags(MT_DrawRGBLine, bitmap, startx, dycoord + Self->Height + 1, dxcoord, dycoord + Self->Height + 1,
                  Self->prvUnderlineRGB.Red, Self->prvUnderlineRGB.Green, Self->prvUnderlineRGB.Blue);
            }

            str++;

            if (Self->AlignFlags & (ALIGN_HORIZONTAL|ALIGN_RIGHT)) {
               tmp       = str;
               strwidth  = 0;
               while ((*tmp) AND (*tmp != '\n')) {
                  tmp += GetUTF8(Self, tmp, &unicode);

                  if (Self->FixedWidth > 0) strwidth += Self->FixedWidth + Self->GlyphSpacing;
                  else {
                     if (unicode < 256) {
                        strwidth += Self->prvChar[unicode].Advance + Self->GlyphSpacing;
                     }
                  }
               }

               if (strwidth) {
                  if (Self->AlignFlags & ALIGN_HORIZONTAL) dxcoord = Self->XCoord + ((Self->AlignWidth - strwidth) / 2);
                  else dxcoord = Self->XCoord + Self->AlignWidth - strwidth;
               }
               else dxcoord = Self->XCoord;
            }
            else dxcoord = Self->XCoord;

            startx = dxcoord;
            dycoord += Self->LineSpacing;
         }
         else if (*str IS '\t') {
            WORD tabwidth;
            tabwidth = (Self->prvChar[' '].Advance + Self->GlyphSpacing) * Self->TabSize;
            dxcoord = Self->XCoord + ROUNDUP(dxcoord - Self->XCoord, tabwidth);
            str++;
         }
         else {
            str += GetUTF8(Self, str, &unicode);

            if (Self->Flags & FTF_CHARCLIP) {
               if (dxcoord + 8 >= Self->WrapWidth) {
                  if (charclip_count++ > 2) break;
                  unicode = '.';
               }
            }

            if (unicode IS ' ') {
               if (Self->FixedWidth > 0) dxcoord += Self->FixedWidth + Self->GlyphSpacing;
               else dxcoord += Self->prvChar[unicode].Advance + Self->GlyphSpacing;
            }
            else {
               if ((unicode > 255) OR (!Self->prvChar[unicode].Advance)) unicode = Self->prvDefaultChar;

               if (Self->Flags & FTF_OUTLINE) {
                  data = Self->prvOutline + Self->prvChar[unicode].OutlineOffset;

                  bytewidth = (Self->prvChar[unicode].Width + 9)>>3;

                  sx = dxcoord - 1;
                  ex = sx + Self->prvChar[unicode].Width + 2;

                  if (ex > bitmap->ClipRight) ex = bitmap->ClipRight;

                  xinc = 0;
                  if (sx < bitmap->ClipLeft) {
                     xinc = bitmap->ClipLeft - sx;
                     sx = bitmap->ClipLeft;
                  }

                  sy = dycoord - 1;
                  if (!(Self->Flags & FTF_DIRECTXY)) sy -= Self->Leading;

                  ey = sy + Self->prvBitmapHeight + 2;
                  if (ey > bitmap->ClipBottom) ey = bitmap->ClipBottom;

                  if (sy < bitmap->ClipTop) {
                     data += bytewidth * (bitmap->ClipTop - sy);
                     sy = bitmap->ClipTop;
                  }

                  sx += bitmap->XOffset;
                  sy += bitmap->YOffset;
                  dx += bitmap->XOffset;
                  dy += bitmap->YOffset;
                  ex += bitmap->XOffset;
                  ey += bitmap->YOffset;

                  for (dy=sy; dy < ey; dy++) {
                     xpos = xinc;
                     for (dx=sx; dx < ex; dx++) {
                        if (data[xpos>>3] & (0x80>>(xpos & 0x7))) {
                           bitmap->DrawUCRPixel(bitmap, dx, dy, &outlinergb);
                        }
                        xpos++;
                     }
                     data += bytewidth;
                  }
               }

               data = Self->prvData + Self->prvChar[unicode].Offset;

               bytewidth = (Self->prvChar[unicode].Width + 7)>>3;

               sx = dxcoord;
               ex = sx + Self->prvChar[unicode].Width;

               if (ex > bitmap->ClipRight) ex = bitmap->ClipRight;

               xinc = 0;
               if (sx < bitmap->ClipLeft) {
                  xinc = bitmap->ClipLeft - sx;
                  sx = bitmap->ClipLeft;
               }

               sy = dycoord;
               if (!(Self->Flags & FTF_DIRECTXY)) sy -= Self->Leading;

               ey = sy + Self->prvBitmapHeight;
               if (ey > bitmap->ClipBottom) ey = bitmap->ClipBottom;

               if (sy < bitmap->ClipTop) {
                  data += bytewidth * (bitmap->ClipTop - sy);
                  sy = bitmap->ClipTop;
               }

               sx += bitmap->XOffset; /* Add offsets only after clipping adjustments */
               sy += bitmap->YOffset;
               dx += bitmap->XOffset;
               dy += bitmap->YOffset;
               ex += bitmap->XOffset;
               ey += bitmap->YOffset;

               if (Self->prvTranslucency < 255) {
                  for (dy=sy; dy < ey; dy++) {
                     xpos = xinc;
                     for (dx=sx; dx < ex; dx++) {
                        if (data[xpos>>3] & (0x80>>(xpos & 0x7))) {
                           bitmap->ReadUCRPixel(bitmap, dx, dy, &rgb);
                           rgb.Red   = Self->prvColourRGB.Red   + (((rgb.Red   - Self->prvColourRGB.Red) * alpha)>>8);
                           rgb.Green = Self->prvColourRGB.Green + (((rgb.Green - Self->prvColourRGB.Green) * alpha)>>8);
                           rgb.Blue  = Self->prvColourRGB.Blue  + (((rgb.Blue  - Self->prvColourRGB.Blue) * alpha)>>8);
                           bitmap->DrawUCRPixel(bitmap, dx, dy, &rgb);
                        }
                        xpos++;
                     }
                     data += bytewidth;
                  }
               }
               else {
                  UBYTE table[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
                  UBYTE *xdata;

                  if (bitmap->BytesPerPixel IS 4) {
                     ULONG *dest;
                     dest = (ULONG *)(bitmap->Data + (sx<<2) + (sy * bitmap->LineWidth));
                     for (dy=sy; dy < ey; dy++) {
                        xpos = xinc & 0x07;
                        xdata = data + (xinc>>3);
                        for (dx=0; dx < ex-sx; dx++) {
                           if (*xdata & table[xpos++]) dest[dx] = getcolour.Colour;
                           if (xpos > 7) {
                              xpos = 0;
                              xdata++;
                           }
                        }
                        dest = (ULONG *)(((UBYTE *)dest) + bitmap->LineWidth);
                        data += bytewidth;
                     }
                  }
                  else if (bitmap->BytesPerPixel IS 2) {
                     UWORD *dest;
                     dest = (UWORD *)(bitmap->Data + (sx<<1) + (sy * bitmap->LineWidth));
                     for (dy=sy; dy < ey; dy++) {
                        xpos = xinc & 0x07;
                        xdata = data + (xinc>>3);
                        for (dx=0; dx < ex-sx; dx++) {
                           if (*xdata & table[xpos++]) dest[dx] = getcolour.Colour;
                           if (xpos > 7) {
                              xpos = 0;
                              xdata++;
                           }
                        }
                        dest = (UWORD *)(((UBYTE *)dest) + bitmap->LineWidth);
                        data += bytewidth;
                     }
                  }
                  else {
                     for (dy=sy; dy < ey; dy++) {
                        xpos = xinc & 0x07;
                        xdata = data + (xinc>>3);
                        for (dx=sx; dx < ex; dx++) {
                           if (*xdata & table[xpos++]) bitmap->DrawUCPixel(bitmap, dx, dy, getcolour.Colour);
                           if (xpos > 7) {
                              xpos = 0;
                              xdata++;
                           }
                        }
                        data += bytewidth;
                     }
                  }
               }

               if (Self->FixedWidth > 0) dxcoord += Self->FixedWidth + Self->GlyphSpacing;
               else dxcoord += Self->prvChar[unicode].Advance + Self->GlyphSpacing;
            }
         }
      }

      /*** Draw an underline for the current line if underlining is turned on ***/

      if (Self->Flags & (FTF_UNDERLINE|FTF_HEAVYLINE)) {
         ActionTags(MT_DrawRGBLine, bitmap, startx, dycoord + Self->Height, dxcoord, dycoord + Self->Height,
            Self->prvUnderlineRGB.Red, Self->prvUnderlineRGB.Green, Self->prvUnderlineRGB.Blue);
      }

      if (Self->Flags & FTF_HEAVYLINE) {
         ActionTags(MT_DrawRGBLine, bitmap, startx, dycoord + Self->Height + 1, dxcoord, dycoord + Self->Height + 1,
            Self->prvUnderlineRGB.Red, Self->prvUnderlineRGB.Green, Self->prvUnderlineRGB.Blue);
      }

      /*** Update X/Y coordinates if requested ***/

      if (Self->Flags & FTF_UPDATEXY) {
         Self->XCoord = dxcoord;
         Self->YCoord = dycoord;
      }

      Action(AC_Unlock, bitmap, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Font: Free
*/

static ERROR FONT_Free(struct Font *Self, APTR Void)
{
   WORD i;

   if (Self->prvData) {
      for (i=0; i < glCacheSize; i++) {
         if (glBitmapCache[i].Data IS Self->prvData) {
            glBitmapCache[i].OpenCount--;

            if (glBitmapCache[i].OpenCount < 0) DPrintF("@Free:","[Font] Deallocation over-run in font cache (index %d).", i);

            if (glBitmapCache[i].OpenCount <= 0) {
               if (glBitmapCache[i].Data)    FreeMemory(glBitmapCache[i].Data);
               if (glBitmapCache[i].Outline) FreeMemory(glBitmapCache[i].Outline);

               ClearMemory(glBitmapCache + i, sizeof(struct BitmapCache));
            }
            break;
         }
      }
   }

   if (Self->FTFace) {
      FT_Done_Face(Self->FTFace);
      Self->FTFace = NULL;
   }

   if ((Self->String) AND (Self->String != Self->prvBuffer)) {
      if (FreeMemory(Self->String) != ERR_Okay) {
         DPrintF("@Free:","[Font] The String field was set illegally (please use SetField)");
      }
      Self->String = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Font: Init
*/

static ERROR FONT_Init(struct Font *Self, APTR Void)
{
   struct mtSelectFont select;
   OBJECTPTR fontserver, file;
   LONG i, diff, abs, error, offset, style;
   WORD index;
   UBYTE alpha, *data;
   FT_Face ftface;
   FT_Open_Args openargs;

   if ((!Self->prvFace[0]) AND (!Self->prvLocation[0])) {
      DPrintF("@Init:","[Font] Face not defined.");
      return(ERR_FieldNotSet);
   }

   if (!Self->Point) Self->Point = 8;

   /*** Search the SystemFonts object to gather initial information about this face ***/

   if ((!Self->Location) OR (!Self->prvLocation[0])) {
      if (AccessObject(glSystemFontsID, 5000, &fontserver) IS ERR_Okay) {
         select.Name  = Self->prvFace;
         select.Style = Self->prvStyle;
         select.Point = Self->Point;
         select.Flags = Self->Flags & (FTF_PREFERSCALED|FTF_PREFERFIXED);
         if (Action(MT_SelectFont, fontserver, &select) IS ERR_Okay) {
            StrCopy(select.Location, Self->prvLocation, SIZE_FONTLOCATION);
            FreeMemory(select.Location);
         }
         else {
            ReleaseObject(fontserver);
            DPrintF("@Init:","[Font] Font \"%s\" (point %d, style %s) is not recognised.", Self->prvFace, Self->Point, Self->prvStyle);
            return(ERR_Failed);
         }

         ReleaseObject(fontserver);
      }
      else {
         DPrintF("@Init:","[Font] Failed to gain access to the font server.");
         return(ERR_ExclusiveDenied);
      }
   }

   /*** Check the bitmap cache to see if we have already loaded this font ***/

   if (StrCompare("Bold", Self->prvStyle, 0, STR_MATCHLENGTH) IS ERR_Okay) style = FTF_BOLD;
   else if (StrCompare("Italic", Self->prvStyle, 0, STR_MATCHLENGTH) IS ERR_Okay) style = FTF_ITALIC;
   else if (StrCompare("Bold Italic", Self->prvStyle, 0, STR_MATCHLENGTH) IS ERR_Okay) style = FTF_BOLD|FTF_ITALIC;
   else style = NULL;

   index = CheckBitmapCache(Self, style);

   if (index != -1) {
      /*** The font exists in the cache ***/
   }
   else if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
      FID_Location|TSTRING, Self->prvLocation,
      FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE,
      TAGEND) IS ERR_Okay) {

      struct winmz_header_fields mz_header;
      struct winne_header_fields ne_header;
      struct winfnt_header_fields header, face;
      struct acRead read;
      ULONG res_offset, font_offset;
      UWORD size_shift, font_count, type_id, count;

      /*** Check if the file is a Windows Bitmap Font ***/

      read.Buffer = (APTR)&mz_header;
      read.Length = sizeof(mz_header);
      Action(AC_Read, file, &read);

      if (mz_header.magic IS ID_WINMZ) {
         ActionTags(AC_Seek, file, (DOUBLE)mz_header.lfanew, POS_START);

         read.Buffer = (APTR)&ne_header;
         read.Length = sizeof(ne_header);
         if ((Action(AC_Read, file, &read) IS ERR_Okay) AND (ne_header.magic IS ID_WINNE)) {

            res_offset = mz_header.lfanew + ne_header.resource_tab_offset;
            ActionTags(AC_Seek, file, (DOUBLE)res_offset, POS_START);

            /*** Count the number of fonts in the file ***/

            font_count  = 0;
            font_offset = 0;
            size_shift  = ReadWordLE(file);

            for (type_id=ReadWordLE(file); type_id; type_id=ReadWordLE(file)) {
               count = ReadWordLE(file);

               if (type_id IS 0x8008) {
                  font_count  = count;
                  GetField(file, FID_Position, FT_LONG, &font_offset);
                  font_offset = font_offset + 4;
                  break;
               }

               ActionTags(AC_Seek, file, (DOUBLE)(4 + count * 12), POS_CURRENT);
            }

            if ((!font_count) OR (!font_offset)) {
               DPrintF("@Init:","[Font] There are no fonts in the file \"%s\"", Self->prvLocation);
               Action(AC_Free, file, NULL);
               return(ERR_Failed);
            }

            ActionTags(AC_Seek, file, (DOUBLE)font_offset, POS_START);

            /*** Scan the list of available fonts to find the closest point size for our font ***/

            struct FontList fonts[font_count];

            for (i=0; i < font_count; i++) {
               fonts[i].Offset = ReadWordLE(file)<<size_shift;
               fonts[i].Size   = ReadWordLE(file)<<size_shift;
               ActionTags(AC_Seek, file, 8.0, POS_CURRENT);
            }

            abs = 0x7fff;
            offset = 0;

            for (i=0; i < font_count; i++) {
               ActionTags(AC_Seek, file, (DOUBLE)fonts[i].Offset, POS_START);

               read.Buffer = (APTR)&header;
               read.Length = sizeof(header);
               if (Action(AC_Read, file, &read) IS ERR_Okay) {
                  if ((header.version != 0x200) AND (header.version != 0x300)) {
                     DPrintF("@Init:","[Font] Font \"%s\" is written in unsupported version %d.", Self->prvFace, header.version);
                     Action(AC_Free, file, NULL);
                     return(ERR_Failed);
                  }

                  if (header.file_type & 1) {
                     DPrintF("@Init:","[Font] Font \"%s\" is in the non-supported vector font format.", Self->prvFace);
                     Action(AC_Free, file, NULL);
                     return(ERR_Failed);
                  }

                  if (header.pixel_width <= 0) header.pixel_width = header.pixel_height;

                  if ((diff = Self->Point - header.nominal_point_size) < 0) diff = -diff;

                  if (diff < abs) {
                     face   = header;
                     abs    = diff;
                     offset = fonts[i].Offset;
                  }
               }
               else {
                  Action(AC_Free, file, NULL);
                  return(ObjectError(Self, ERH_Init, ERR_Read));
               }
            }

            /* Scan the bitmap cache again to ensure that the discovered font is
            ** not already loaded.  This is important if the cached font wasn't originally
            ** found due to slight differences in point size.
            */

            Self->Point = face.nominal_point_size;
            index = CheckBitmapCache(Self, style);
            if (index IS -1) {
               /*** Load the font into the cache ***/

               index = CacheBitmapFont(Self, file, offset, &face);
               if (index IS -1) {
                  Action(AC_Free, file, NULL);
                  return(ERR_Failed);
               }
            }

         } /* File is not a windows fixed font (but could be truetype) */
      } /* File is not a windows fixed font (but could be truetype) */

      Action(AC_Free, file, NULL);
   }
   else return(ObjectError(Self, ERH_Init, ERR_Read));

   if (index != -1) {

      Self->prvData     = glBitmapCache[index].Data;
      Self->prvOutline  = glBitmapCache[index].Outline;
      Self->Point       = glBitmapCache[index].Header.nominal_point_size;
      Self->Height      = glBitmapCache[index].Header.ascent - glBitmapCache[index].Header.internal_leading + glBitmapCache[index].Header.external_leading;
      Self->Leading     = glBitmapCache[index].Header.internal_leading;
      Self->Gutter      = glBitmapCache[index].Header.external_leading;
      Self->LineSpacing += glBitmapCache[index].Header.pixel_height; /* Add to any preset linespacing rather than over-riding */
      Self->MaxHeight   = glBitmapCache[index].Header.pixel_height; /* Supposedly the pixel_height includes internal and external leading values (?) */
      Self->prvBitmapHeight = glBitmapCache[index].Header.pixel_height;
      Self->prvDefaultChar  = glBitmapCache[index].Header.first_char + glBitmapCache[index].Header.default_char;
      Self->TotalCharacters = glBitmapCache[index].Header.last_char - glBitmapCache[index].Header.first_char + 1;

      if (Self->FixedWidth > 0) Self->prvSpaceWidth = Self->FixedWidth;
      else if (glBitmapCache[index].Chars[' '].Advance) Self->prvSpaceWidth = glBitmapCache[index].Chars[' '].Advance;
      else Self->prvSpaceWidth = glBitmapCache[index].Chars[glBitmapCache[index].Header.first_char + glBitmapCache[index].Header.break_char].Advance;

      DPrintF("Font:","Index: %d, Count: %d, Style: %s", index, glBitmapCache[index].OpenCount, Self->prvStyle);

      if (!glBitmapCache[index].OpenCount) {
         /* Modify the font characters to bold and/or italic text if requested and
         ** the font file is in the regular style.
         */

         if ((StrCompare("Bold", Self->prvStyle, 0, STR_MATCHLENGTH) IS ERR_Okay) AND (glBitmapCache[index].Header.weight < 600)) {
            if (ConvertFontGraphic(Self, glBitmapCache + index, FTF_BOLD, &data) != ERR_Okay) return(ERR_Failed);
            glBitmapCache[index].StyleFlags |= FTF_BOLD;

            FreeMemory(glBitmapCache[index].Data);
            glBitmapCache[index].Data = data;
            Self->prvData = data;
         }
         else if ((StrCompare("Italic", Self->prvStyle, 0, STR_MATCHLENGTH) IS ERR_Okay) AND (!glBitmapCache[index].Header.italic)) {
            if (ConvertFontGraphic(Self, glBitmapCache + index, FTF_ITALIC, &data) != ERR_Okay) return(ERR_Failed);
            glBitmapCache[index].StyleFlags |= FTF_ITALIC;

            FreeMemory(glBitmapCache[index].Data);
            glBitmapCache[index].Data = data;
            Self->prvData = data;
         }
         else if ((StrCompare("Bold Italic", Self->prvStyle, 0, STR_MATCHLENGTH) IS ERR_Okay) AND
                  (!glBitmapCache[index].Header.italic) AND (glBitmapCache[index].Header.weight < 600)) {
            if (ConvertFontGraphic(Self, glBitmapCache + index, FTF_BOLD|FTF_ITALIC, &data) != ERR_Okay) return(ERR_Failed);
            glBitmapCache[index].StyleFlags |= FTF_BOLD|FTF_ITALIC;

            FreeMemory(glBitmapCache[index].Data);
            glBitmapCache[index].Data = data;
            Self->prvData = data;
         }

      }

      if (Self->Flags & FTF_OUTLINE) {
         if (!glBitmapCache[index].Outline) GenerateFontOutline(Self, glBitmapCache + index, &glBitmapCache[index].Outline);

         Self->prvOutline = glBitmapCache[index].Outline;
      }

      CopyMemory(glBitmapCache[index].Chars, Self->prvChar, sizeof(glBitmapCache[index].Chars));

      Self->Flags |= glBitmapCache[index].StyleFlags;

      /*** Finally we perform word wrapping ***/

      if (Self->Flags & FTF_WORDWRAP) DoWordWrap(Self);

      glBitmapCache[index].OpenCount++;

      return(ERR_Okay);
   }

   /*** Attempt to load a truetype font file ***/

   openargs.flags    = FT_OPEN_PATHNAME;
   openargs.pathname = Self->prvLocation;
   if ((error = FT_Open_Face(glFTLibrary, &openargs, 0, &ftface))) {
      if (error IS FT_Err_Unknown_File_Format) {
         return(ERR_NoSupport);
      }

      DPrintF("@Init:","[Font] Fatal error in attempting to load font \"%s\".", Self->prvLocation);
      return(ERR_Failed);
   }

   if (FT_HAS_KERNING(ftface)) Self->Flags |= FTF_KERNING;

   if (FT_IS_SCALABLE(ftface)) Self->Flags |= FTF_SCALABLE;
   else {
      /*** Only scalable fonts are supported by this routine ***/
      FT_Done_Face(ftface);
      return(ObjectError(Self, ERH_Init, ERR_NoSupport));
   }

   /* Choose the point size that most closely matches the
   ** requested point size.
   */

   FT_Set_Char_Size(ftface, 0, Self->Point<<FT_DOWNSIZE, SCREEN_DPI, SCREEN_DPI);

   Self->FTFace = ftface;
   Self->TotalCharacters = ftface->num_glyphs;

   /* Calculate the Height.  For scalable fonts this is calculated by
   ** scanning characters A - Z and determining which is the biggest
   ** character of those that have been analysed.
   */

   if (!Self->Height) {
      for (alpha='A'; alpha <= 'Z'; alpha++) {
         if (!FT_Load_Glyph(ftface, FT_Get_Char_Index(Self->FTFace, alpha), FT_LOAD_DEFAULT)) {
            i = ftface->glyph->metrics.horiBearingY>>FT_DOWNSIZE;
            if (i > Self->Height) Self->Height = i;
         }
      }

      for (alpha='a'; alpha <= 'z'; alpha++) {
         if (!FT_Load_Glyph(ftface, FT_Get_Char_Index(Self->FTFace, alpha), FT_LOAD_DEFAULT)) {
            i = ftface->glyph->metrics.horiBearingY>>FT_DOWNSIZE;
            if (i > Self->Height) Self->Height = i;
         }
      }

      if (!Self->Height) Self->Height = Self->Point;
   }

   /* Determine the line distance of the font, which describes the amount of distance
   ** between each font line that is printed.
   */

   Self->LineSpacing += Self->Height * 1.33;

   Self->MaxHeight = Self->Height * 1.33;

   /*** Leading adjustments for the top part of the font ***/

   Self->Leading = Self->MaxHeight - Self->Height; /* Make the leading the same size as the gutter */
   Self->MaxHeight   += Self->Leading; /* Increase the max-height by the leading amount */
   Self->LineSpacing += Self->Leading; /* Increase the line-spacing by the leading amount */

   if (Self->FixedWidth > 0) Self->prvSpaceWidth = Self->FixedWidth;
   else {
      if (!FT_Load_Glyph(ftface, FT_Get_Char_Index(Self->FTFace, FT_Get_Char_Index(ftface, CHAR_SPACE)), FT_LOAD_DEFAULT)) {
         Self->prvSpaceWidth = (ftface->glyph->advance.x>>FT_DOWNSIZE);
         if (Self->prvSpaceWidth < 3) Self->prvSpaceWidth = Self->Height * 0.5;
      }
      else Self->prvSpaceWidth = Self->Height * 0.5;
   }

   /*** Precalculate the width of each character in the range of 0x20 - 0xff ***/

   for (i=' '; i < 256; i++) {
      if (!FT_Load_Glyph(ftface, FT_Get_Char_Index(ftface, i), FT_LOAD_DEFAULT)) {
         Self->prvChar[i].Width = ftface->glyph->advance.x>>FT_DOWNSIZE;
         Self->prvChar[i].Advance = ftface->glyph->advance.x>>FT_DOWNSIZE;
      }
   }

   /*** Turn on anti-aliasing if the font is scalable ***/

   if ((Self->Flags & FTF_SCALABLE) AND (!(Self->Flags & FTF_QUICKALIAS))) Self->Flags |= FTF_ANTIALIAS;

   /*** Turn off outlining if aliasing is on ***/

   if ((Self->Flags & FTF_OUTLINE) AND (Self->Flags & (FTF_QUICKALIAS|FTF_ANTIALIAS))) {
      Self->Flags = Self->Flags & (~FTF_OUTLINE);
   }

   /*** Finally we perform word wrapping ***/

   if (Self->Flags & FTF_WORDWRAP) DoWordWrap(Self);

   DPrintF("Init:","[Font] Family: %s, Style: %s, Glyphs: %d, Point: %d, Height: %d", Self->prvFace, Self->prvStyle, Self->TotalCharacters, Self->Point, Self->Height);
   //DPrintF("Init:","[Font] LineSpacing: %d, Leading: %d, Gutter: %d", Self->LineSpacing, Self->Leading, Self->Gutter);

   return(ERR_Okay);
}

/*****************************************************************************
** Font: NewObject
*/

static ERROR FONT_NewObject(struct Font *Self, APTR Void)
{
   Self->ColourRGB       = &Self->prvColourRGB;
   Self->UnderlineRGB    = &Self->prvUnderlineRGB;
   Self->OutlineRGB      = &Self->prvOutlineRGB;
   Self->TabSize         = 8;
   Self->prvTranslucency = 255;
   Self->WrapWidth       = 50;
   Self->prvDefaultChar  = '.';
   Self->Location        = Self->prvLocation;
   Self->Style           = Self->prvStyle;
   Self->Face            = Self->prvFace;
   StrCopy("Regular", Self->prvStyle, sizeof(Self->prvStyle));
   return(ERR_Okay);
}

/*****************************************************************************
** Function: ReadByte()
** Short:    Reads any double digit hex value and converts it into a number.
*/

static WORD ReadByte(STRING str)
{
   WORD value;
   if (!*str) return(0);
   if ((*str >= '0') AND (*str <= '9')) value = (*str - '0') * 16;
   else if ((*str >= 'A') AND (*str <= 'F')) value = ((*str - 'A')+10) * 16;
   else if ((*str >= 'a') AND (*str <= 'f')) value = ((*str - 'a')+10) * 16;
   else value = 0;
   str++;
   if ((*str >= '0') AND (*str <= '9')) value += (*str - '0');
   else if ((*str >= 'A') AND (*str <= 'F')) value += (*str - 'A') + 10;
   else if ((*str >= 'a') AND (*str <= 'f')) value += (*str - 'a') + 10;
   return(value);
}

/*****************************************************************************

###FIELD###
Name:   Align
Short:  Sets the position of a font string to an abstract alignment.
Type:   LONG
Prefix: ALIGN
Status: Set

###DESCRIPTION###
<p>You can use this field to set the alignment of a font within its drawable
area.  This is an abstract means of positioning in comparison to setting the
XCoord and YCoord fields directly.  Available alignment flags are listed in
the following table:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">BOTTOM</td><td>Aligns the font to the bottom of the drawable area.</td></tr>
<tr><td width="10%">CENTER</td><td>Centers the font on both the vertical and horizontal axis.</td></tr>
<tr><td width="10%">LEFT</td><td>Aligns the font to the left (horizontal coordinate 0).</td></tr>
<tr><td width="10%">MIDDLE</td><td>Synonym for the CENTER flag.</td></tr>
<tr><td width="10%">RIGHT</td><td>Aligns the font so that the last character is positioned flush against the right.</td></tr>
<tr><td width="10%">TOP</td><td>Aligns the font to the top (vertical coordinate 0).</td></tr>
<tr><td width="10%">HORIZONTAL</td><td>Centers the font on the horizontal axis.</td></tr>
<tr><td width="10%">VERTICAL</td><td>Centers the font on the vertical axis.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Bitmap
Short:  The destination Bitmap to use when drawing a font.
Type:   struct Bitmap *
Status: Read/Set

###DESCRIPTION###
<p>Before calling the Draw action on a font object, the Bitmap field must be
set so that the font has a graphical destination.  This field is dynamic, so
a font does not have to be anchored to a Bitmap after initialisation (i.e.
you can use a single font object to draw to as many different Bitmap types
as you please).</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Bold
Short:  Set to TRUE to enable bold styling.
Type:   LONG
Status: Get/Init

###DESCRIPTION###
<p>You may set the Bold field to TRUE prior to initialisation if you would
like a font to use bold styling.  This field is provided only for convenience
- we recommend that you set the Style field for determining font styling where
possible.</p>

###SEE ALSO###
Field: Style

###END###

*****************************************************************************/

static ERROR GET_Bold(struct Font *Self, LONG *Value)
{
   if (Self->Flags & FTF_BOLD) *Value = TRUE;
   else if (StrSearch("bold", Self->prvStyle, NULL) != -1) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Bold(struct Font *Self, LONG *Value)
{
   return(SET_Style(Self, "Bold"));
}

/*****************************************************************************

###FIELD###
Name:   Colour
Short:  String-based field for defining the colour of a font.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The colour of a font can be declared by writing to this field.  The
colour must be in hexadecimal or separated-decimal format - for example
to create a pure red colour, a setting of &quot;#ff0000&quot; or
&quot;255,0,0&quot; would be valid.</p>

<p>If you create a font object without setting the Colour field, the text will
be printed in black.</p>

###SEE ALSO###
Field: ColourRGB

###END###

*****************************************************************************/

static ERROR SET_Colour(struct Font *Self, STRING Colour)
{
   if ((!Colour) OR (!Colour[0])) return(ERR_Okay);

   if (*Colour IS '#') {
      Colour++;
      Self->ColourRGB->Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->ColourRGB->Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->ColourRGB->Blue  = ReadByte(Colour);
   }
   else {
      Self->ColourRGB->Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->ColourRGB->Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->ColourRGB->Blue = StrToInt(Colour);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ColourRGB
Short:  The base colour of a font.
Type:   RGB
Status: Read/Set

###DESCRIPTION###
<p>The base RGB colour of a Font is defined in this field.  If the colour is
not set then the Font will be drawn in black text.</p>

###END###

*****************************************************************************/

static ERROR SET_ColourRGB(struct Font *Font, struct RGB *RGB)
{
   if (Font->ColourRGB) {
      Font->ColourRGB->Red   = RGB->Red;
      Font->ColourRGB->Green = RGB->Green;
      Font->ColourRGB->Blue  = RGB->Blue;
      return(ERR_Okay);
   }
   else return(ERR_Failed);
}

/*****************************************************************************

###FIELD###
Name:   Face
Short:  The name of a font face that is to be loaded on initialisation.
Type:   STRING
Status: Read/Init

###DESCRIPTION###
<p>The name of the face that you wish to use for a font must be specified
here.  If this field is not set then the initialisation process will use the
user's preferred face.  A list of available faces can be obtained from
the 'SystemFonts' public object.</p>

<p>For convenience, the face string can also be extended with extra parameters so
that you can define point size and style information at the same time when
writing this field.  Extra parameters are delimited with the colon character
and must follow a set order as defined here:</p>

<pre>
   face:pointsize:style:colour
</pre>

<p>Here are some examples:</p>

<pre>
  Helvetica:12:Bold Italic:#ff0000
  Courier:10
  Charter:16::255,128,255
</pre>

<p>To load a font file that is not installed by default, replace the face
parameter with the SRC command, followed by the font location:</p>

<pre>
  SRC:exodus:data/images/shine:14:Italic
</pre>

###END###

*****************************************************************************/

static ERROR SET_Face(struct Font *Self, STRING Name)
{
   LONG i, j, k;

   if ((Name) AND (Name[0])) {
      if (StrCompare("SRC:", Name, 4, NULL) IS ERR_Okay) {
         j = 0;
         for (i=4,k=0; Name[i]; i++) {
            if (Name[i] IS ':') {
               j++;
               if (j > 1) break;
            }
            Self->prvLocation[k++] = Name[i];
         }
         Self->prvLocation[k] = 0;
         Self->prvFace[0] = 0;
      }
      else {
         for (i=0; (Name[i]) AND (Name[i] != ':') AND (i < SIZE_FONTFACE-1); i++) Self->prvFace[i] = Name[i];
         Self->prvFace[i] = 0;
      }

      if (Name[i] != ':') return(ERR_Okay);

      /*** Extract the point size ***/

      i++;
      Self->Point = StrToInt(Name+i);

      while ((Name[i] >= '0') AND (Name[i] <= '9')) i++;

      if (Name[i] != ':') return(ERR_Okay);

      /*** Extract the style string ***/

      i++;
      for (j=0; (Name[i]) AND (Name[i] != ':') AND (j < SIZE_FONTSTYLE-1); j++) Self->prvStyle[j] = Name[i++];
      Self->prvStyle[j] = 0;

      if (Name[i] != ':') return(ERR_Okay);

      /*** Extract the colour string ***/

      i++;
      SET_Colour(Self, Name + i);
   }
   else Self->prvFace[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags.
Type:   LONG
Prefix: FTF
Status: Read/Init

###DESCRIPTION###
<p>This field contains flags that affect font behaviour. Currently available
flags are:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">ANTIALIAS</td><td>Smooths text with an anti-alias technique.  Anti-aliasing is forced automatically when scalable fonts are used.  This flag has no purpose when using fixed bitmap fonts.</td></tr>
<tr><td width="10%">BOLD</td><td>Indicates that bold text is enabled for the font (read only).</td></tr>
<tr><td width="10%">HEAVYLINE</td><td>Enables font underlining at twice the normal amount, using the colour in the <a href="#UnderlineRGB">UnderlineRGB</a> field.</td></tr>
<tr><td width="10%">ITALICS</td><td>Indicates that italic text is enabled for the font (read only).</td></tr>
<tr><td width="10%">OUTLINE</td><td>Enables font outlines using the colour in the <a href="#OutlineRGB">OutlineRGB</a> field.  Only works with bitmap fonts.</td></tr>
<tr><td width="10%">UNDERLINE</td><td>Enables font underlining using the colour in the <a href="#UnderlineRGB">UnderlineRGB</a> field.</td></tr>
<tr><td width="10%">UPDATEXY</td><td>If this field is set then the XCoord and YCoord fields will be increase to the next cursor position after a successful Draw action.</td></tr>
<tr><td width="10%">WORDWRAP</td><td>Wraps text back to the value given in a font's <a href="#XCoord">XCoord</a> every time the edge of the drawing area is reached.</td></tr>
</table>

###END###

*****************************************************************************/

static ERROR SET_Flags(struct Font *Self, LONG *Flags)
{
   Self->Flags = (Self->Flags & 0xff000000) | (*Flags & 0x00ffffff);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Height
Short:  The full height of a font (distance between font lines).
Type:   LONG
Status: Read

###DESCRIPTION###
<p>The height of a font object is defined in this field and describes the
distance between individual font lines (measured in pixels).  The font height
is calculated on initialisation and can be read at any time.</p>

###SEE ALSO###
Field: Point

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   GlyphSpacing
Short:  The amount of spacing between each character.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field represents the horizontal spacing between each glyph,
technically known as kerning between each font character.  Fonts that have a
high GlyphSpacing value will typically print out like this:</p>

<pre>
     H e l l o   W o r l d !
</pre>

<p>On the other hand, using negative values in this field can cause text
to be printed backwards.  The GlyphSpacing value is typically set to zero or one
by default, depending on the font type that has been loaded.</p>

###SEE ALSO###
Field: LineSpacing

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   LineCount
Short:  The amount of lines in a font String.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>You can read this field if you need to know how many lines are present in
a font's String field.  This figure does not include lines that may result
from column wrapping, i.e. only specific line feeds are recognised.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   LineSpacing
Short:  The amount of spacing between each line.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field defines the amount of space between each line that is printed
with a font object.  It is set automatically during initialisation to reflect
the recommended distance between each line.   You can increase or decrease
this value to make finer adjustments to the line spacing.  If negative, the
text will be printed in a reverse vertical direction with each new line.</p>

<p>If you set this field prior to initialisation, the value that you set will
be added to the font's normal line-spacing, instead of over-riding it.  For
instance, setting the LineSpacing to 2 will result in an extra 2 pixels being
added to the font's spacing.</p>

###SEE ALSO###
Field: GlyphSpacing

###END###

*****************************************************************************/

/*****************************************************************************

*****************************************************************************/

static ERROR SET_Location(struct Font *Self, STRING Value)
{
   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) {
      if (Value) {
         Self->Location = Self->prvLocation;
         StrCopy(Value, Self->prvLocation, SIZE_FONTLOCATION);
      }
      else {
         Self->Location = NULL;
         Self->prvLocation[0] = 0;
      }
      return(ERR_Okay);
   }
   else return(ERR_Failed);
}

/*****************************************************************************

###FIELD###
Name:    Opacity
Synonym: Translucency
Short:   Determines the level of translucency applied to a font.
Type:    FLOAT
Status:  Get/Set

###DESCRIPTION###
<p>This field determines the translucency level of a font graphic.  The
default setting is 100%, which means that the font will not be translucent.  Any
other value that you set here will alter the impact of a font's graphic over
the destination Bitmap.  High values will retain the boldness of the font,
while low values can render it close to invisible.</p>

<p>Please note that the use of translucency will always have an impact on the
time it normally takes to draw a font.</p>

###END###

*****************************************************************************/

static ERROR GET_Opacity(struct Font *Self, FLOAT *Value)
{
   *Value = Self->prvTranslucency * 100 / 255;
   return(ERR_Okay);
}

static ERROR SET_Opacity(struct Font *Self, FLOAT *Value)
{
   Self->Flags |= FTF_ANTIALIAS;
   if (*Value >= 100) Self->prvTranslucency = 0;
   else if (*Value <= 0) Self->prvTranslucency = 255;
   else Self->prvTranslucency = *Value * 255 / 100;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Outline
Short:  String-based field for defining the outline colour of a font.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>If you want to use font outlining, you need to set either the Outline or
OutlineRGB field so that the font object knows what colour to use for the
outline.  When writing to this field, the colour must be declared in
hexadecimal or separated-decimal format.  For example to create a pure red
colour, a setting of &quot;#ff0000&quot; or &quot;255,0,0&quot; would be
valid.</p>

###SEE ALSO###
Field: OutlineRGB

###END###

*****************************************************************************/

static ERROR SET_Outline(struct Font *Self, STRING Colour)
{
   if (!Colour) {
      Self->Flags &= ~FTF_OUTLINE;
      return(ERR_Okay);
   }
   else if (*Colour IS '#') {
      Colour++;
      Self->OutlineRGB->Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->OutlineRGB->Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->OutlineRGB->Blue  = ReadByte(Colour);
   }
   else {
      Self->OutlineRGB->Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->OutlineRGB->Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->OutlineRGB->Blue = StrToInt(Colour);
   }

   Self->Flags |= FTF_OUTLINE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   OutlineRGB
Short:  The colour of a font's outline.
Type:   RGB
Status: Set

###DESCRIPTION###
<p>If you have used the OUTLINE <a href="#Flags">Flag</a> then you may want to
set the OutlineRGB field if you would like an outline colour other than
black.</p>

###END###

*****************************************************************************/

static ERROR SET_OutlineRGB(struct Font *Font, struct RGB *RGB)
{
   if (Font->OutlineRGB) {
      Font->OutlineRGB->Red   = RGB->Red;
      Font->OutlineRGB->Green = RGB->Green;
      Font->OutlineRGB->Blue  = RGB->Blue;
      return(ERR_Okay);
   }
   else return(ERR_Failed);
}

/*****************************************************************************

###FIELD###
Name:   Point
Short:  The point size of a font.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>The point size of a font defines the size of a font relative to other point
sizes for a particular font face.  For example, Arial point 8 is half the size
of Arial point 16.  The point size between font families cannot be compared
accurately due to designer discretion when it comes to determining font size.
For accurate point size in terms of pixels, please refer to the Height field.</p>

<p>When setting the point size of a bitmap font, the system will try and find the
closest matching value for the requested point size.  For instance, if you request
a fixed font at point 11 and the closest size is point 8, the system will drop
the font to point 8.  This does not impact upon scalable fonts, which can be
measured to any point size.</p>

###SEE ALSO###
Field: Height

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   String
Short:  The string to use when drawing a Font.
Type:   STRING
Status: Read/Set

###DESCRIPTION###
<p>This field is used when you want to use the Draw action to draw a string
to a drawable area.  A string may consist of any characters that you like and
line feeds are allowed (whenever a line feed is reached, the Draw action will
start printing on the next line).  Drawing will stop when the null termination
character is reached.</p>

<p>If a string contains characters that are not supported by a font, those
characters will be printed using simply be ignored when printing.</p>

###END###

*****************************************************************************/

static ERROR SET_String(struct Font *Self, STRING Value)
{
   LONG i;

   if (StrCompare(Value, Self->String, 0, STR_MATCHCASE|STR_MATCHLENGTH) IS ERR_Okay) return(ERR_Okay);

   if ((Self->String) AND (Self->String != Self->prvBuffer)) {
      FreeMemory(Self->String);
   }

   Self->String      = NULL;
   Self->LineCount   = 0;
   Self->prvStrWidth = 0; /* Reset the string width for GET_Width */

   if ((Value) AND (*Value)) {
      /*** Get the string's byte length ***/
      for (i=0; Value[i]; i++);

      if (i < SIZE_FONTBUFFER-1) {
         /*** Use the internal buffer rather than allocating a memory block ***/
         Self->String = Self->prvBuffer;
         for (i=0; Value[i]; i++) Self->String[i] = Value[i];
         Self->String[i] = 0;
      }
      else if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->String, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->String[i] = Value[i];
         Self->String[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));

      if (Self->Flags & FTF_WORDWRAP) DoWordWrap(Self);

      CalcLineCount(Self);

      /*** Recalculate the pixel width of the string ***/
      if (Self->Head.ObjectFlags & NF_INITIALISED) GET_Width(Self, &Self->prvStrWidth);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Style
Short:  Determines font styling.
Type:   STRING
Status: Read/Init

###DESCRIPTION###
<p>The style of a font can be selected by setting the Style field.  This comes
into effect only if the font actually supports the specified style as part of
its graphics set.  If the style is unsupported, the regular styling of the face
will be used on initialisation.</p>

<p>Bitmap fonts are a special case if a bold or italic style is selected.  In
this situation the system can automatically convert the font to that style
even if the correct graphics set does not exist.</p>

<p>Typical font styles are "Bold", "Bold Italic", "Italic" and "Regular" (the
default).  TrueType fonts can consist of any style that the designer chooses,
such as "Narrow" or "Wide", so check the SystemFonts object if you need to
analyse available styles.

###END###

*****************************************************************************/

static ERROR SET_Style(struct Font *Self, STRING Value)
{
   if ((!Value) OR (!Value[0])) {
      StrCopy("Regular", Self->prvStyle, sizeof(Self->prvStyle));
   }
   else StrCopy(Value, Self->prvStyle, sizeof(Self->prvStyle));

   return(ERR_Okay);
}

/*****************************************************************************

*****************************************************************************/

static ERROR SET_Tabs(struct Font *Self, WORD *Tabs)
{
   LONG i;

   if (!Tabs) return(ERR_Failed);

   for (i=0; Tabs[i]; i++);
   if (AllocMemory(sizeof(WORD) * i, MEM_NOCLEAR, &Self->Tabs, NULL) IS ERR_Okay) {
      CopyMemory(Tabs, Self->Tabs, sizeof(WORD) * i);
      return(ERR_Okay);
   }
   else return(ERR_AllocMemory);
}

/*****************************************************************************

###FIELD###
Name:    TabSize
Short:   Defines the tab size to use when drawing and manipulating a font string.
Type:    LONG
Status:  Read/Write

###DESCRIPTION###
<p>You can set the TabSize of a font to alter the standard tab setting.  The
default tab size is 8, so if you consider this to be too large, then you may
want to shrink it.  The TabSize only comes into effect when tabs are used
in the font <a href="#String">String</a>.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Underline
Short:  String-based field for defining the underlining colour of a font.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>If you want to underline a font graphic, you need to set either the
Underline or UnderlineRGB field so that the font object knows what colour to
use for the underline.  When writing to this field, the colour must be
declared in hexadecimal or separated-decimal format.  For example to create a
pure red colour, a setting of &quot;#ff0000&quot; or &quot;255,0,0&quot; would
be valid.</p>

###SEE ALSO###
Field: UnderlineRGB

###END###

*****************************************************************************/

static ERROR SET_Underline(struct Font *Self, STRING Colour)
{
   if (!Colour) {
      Self->Flags &= ~FTF_UNDERLINE;
      return(ERR_Okay);
   }
   else if (*Colour IS '#') {
      Colour++;
      Self->UnderlineRGB->Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->UnderlineRGB->Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->UnderlineRGB->Blue  = ReadByte(Colour);
   }
   else {
      Self->UnderlineRGB->Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->UnderlineRGB->Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->UnderlineRGB->Blue = StrToInt(Colour);
   }

   Self->Flags |= FTF_UNDERLINE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   UnderlineRGB
Short:  Defines the colour to use when underlining a font.
Type:   *RGB
Status: Read/Set

###DESCRIPTION###
<p>If a font has the underline mode turned on, you can specify the colour
of the line here.  If a colour is not specified here then a black line will
be used.</p>

###END###

*****************************************************************************/

static ERROR SET_UnderlineRGB(struct Font *Font, struct RGB *RGB)
{
   if (Font->UnderlineRGB) {
      Font->UnderlineRGB->Red   = RGB->Red;
      Font->UnderlineRGB->Green = RGB->Green;
      Font->UnderlineRGB->Blue  = RGB->Blue;
      return(ERR_Okay);
   }
   else return(ERR_Failed);
}

/*****************************************************************************

###FIELD###
Name:     Width
Short:    Returns the pixel width of a string.
Type:     LONG
Status:   Get

###DESCRIPTION###
<p>You can read this virtual field to obtain the pixel width of a font
string.  You must have a string already specified in the font for
this to work, or a width of zero will be returned.</p>

###SEE ALSO###
Method: CalcStringWidth

###END###

*****************************************************************************/

static ERROR GET_Width(struct Font *Self, LONG *Value)
{
   struct mtCalcStringWidth calc;

   if (!Self->String) {
      *Value = 0;
      return(ERR_Okay);
   }

   /* NB: The width is only calculated if the prvStrWidth field is
   ** not set.  (This is dependent on the Font->String field, refer
   ** to it for further details).  The reason is speed - caching
   ** the width saves us from recalculating it all the time.
   */

   if (!Self->prvStrWidth) {
      calc.String = Self->String;
      calc.Length = -1;
      Action(MT_CalcStringWidth, Self, &calc);
      *Value = calc.Width;
   }
   else *Value = Self->prvStrWidth;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal coordinate of a font's string.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>When drawing font strings, the XCoord and YCoord fields define the position
that the string will placed at within its drawable container.  If you do not
initialise the coordinates then printing will always start from the top left
of the drawing area.</p>

###SEE ALSO###
Field: YCoord

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical coordinate of a font's string.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>When drawing font strings, the XCoord and YCoord fields define the position
that the string will placed at within its drawable container.  If you do not
initialise the coordinates then printing will always start from the top left
of the drawing area.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  Additional offset value that is added to vertically aligned fonts.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>Fonts that are aligned vertically (either in the center or bottom edge of
the drawing area) will have a vertical offset value.  You can read that
value from this field and add it to the YCoord field in order to get a more
accurate reading when determining where a font will be drawn.</p>

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Font *Self, LONG *Value)
{
   LONG offset, linecount;

   if ((linecount = Self->LineCount) < 1) linecount = 1;

   if (Self->AlignFlags & ALIGN_VERTICAL) {
      offset = (Self->AlignHeight - (Self->Height + (Self->LineSpacing * (linecount-1)))) / 2;
      offset += (Self->LineSpacing - Self->MaxHeight) / 2; /* Adjust for spacing between each individual line */
      if (Self->Flags & FTF_DIRECTXY) offset -= Self->Leading;
      *Value = offset;
   }
   else if (Self->AlignFlags & ALIGN_BOTTOM) {
      *Value = Self->AlignHeight - (Self->MaxHeight + (Self->LineSpacing * (linecount-1)));
      //if (Self->Flags & FTF_DIRECTXY) dycoord -= Self->Leading;
   }
   else *Value = 0;

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CalcLineCount()
*/

static void CalcLineCount(struct Font *Self)
{
   STRING str;

   if ((str = Self->String)) {
      Self->LineCount = 1;
      while (*str) {
         if (*str IS '\n') Self->LineCount++;
         str++;
      }
   }
}
/*****************************************************************************
** Internal: GetKerning()
** Short:    Only call this function if the font includes kerning support (test
**           via FTF_KERNING).
*/

static LONG GetKerning(FT_Face Face, LONG Glyph, LONG PrevGlyph)
{
   FT_Vector delta;

   if ((!Glyph) OR (!PrevGlyph)) return(0);

   FT_Get_Kerning(Face, PrevGlyph, Glyph, FT_KERNING_DEFAULT, &delta);
   return(delta.x>>FT_DOWNSIZE);
}

/*****************************************************************************
** Internal: DoWordWrap()
*/

static void DoWordWrap(struct Font *Self)
{
   STRING str, word;
   LONG x, glyph, prevglyph, unicode;

   if (!Self->String) return;

   str       = Self->String;
   word      = NULL;
   x         = 0;
   prevglyph = 0;
   glyph     = 0;
   while (*str) {
      if (*str IS ' ') {
         while (*str IS ' ') {
            x += Self->prvChar[' '].Advance;
            str++;
         }
         if (!*str) break;
         word = str; /* Remember the starting character of the next word */
      }

      if (*str IS '\n') {
         str++;
         x = 0;
         word = NULL;
      }
      else {
         str += GetUTF8(Self, str, &unicode);

         if (Self->FixedWidth > 0) x += Self->FixedWidth;
         else if (Self->Flags & FTF_SCALABLE) {
            glyph = FT_Get_Char_Index(Self->FTFace, unicode);

            if (Self->Flags & FTF_KERNING) x += GetKerning(Self->FTFace, glyph, prevglyph);

            if (!FT_Load_Glyph(Self->FTFace, glyph, FT_LOAD_DEFAULT)) {
               x += (Self->FTFace->glyph->advance.x>>FT_DOWNSIZE);
            }

            prevglyph = glyph;
         }
         else if (unicode < 256) x += Self->prvChar[unicode].Advance;

         if ((x >= Self->WrapWidth) AND (word)) {
            word[-1] = '\n';
            x = 0;
            str = word; /* Restart the calculation from the return code */
            word = NULL;
         }
         else x += Self->GlyphSpacing;
      }
   }

   /*** Recalculate the number of lines ***/

   CalcLineCount(Self);
}

/*****************************************************************************
** Internal: ConvertFontGraphic()
** Short:    Converts regular bitmap fonts into bold / italic bitmap fonts.
*/

static ERROR ConvertFontGraphic(struct Font *Self, struct BitmapCache *Cache, LONG Flags, UBYTE **Data)
{
   UBYTE *buffer, *gfx, *dest, *fontdata;
   LONG size, pos;
   WORD extra, i, y, xb, oldwidth, newwidth, dx, sx, italic;

   *Data = NULL;
   fontdata = Cache->Data;

   if (Flags & FTF_BOLD) {
      DPrintF("Init:","Converting font graphic to bold.");

      size = 0;
      for (i=0; i < 256; i++) {
         if (Cache->Chars[i].Width) size += Cache->Header.pixel_height * ((Cache->Chars[i].Width+8)>>3);
      }

      if (AllocMemory(size, MEM_UNTRACKED, &buffer, NULL) IS ERR_Okay) {
         pos = 0;
         for (i=0; i < 256; i++) {
            if (Cache->Chars[i].Width) {
               gfx = fontdata + Cache->Chars[i].Offset;
               Cache->Chars[i].Offset = pos;

               /* Copy character graphic to the buffer and embolden it */

               oldwidth = (Cache->Chars[i].Width+7)>>3;
               newwidth = (Cache->Chars[i].Width+8)>>3;
               for (y=0; y < Cache->Header.pixel_height; y++) {
                  for (xb=0; xb < oldwidth; xb++) {
                     buffer[pos+xb] |= gfx[xb]|(gfx[xb]>>1);
                     if ((xb < newwidth) AND (gfx[xb] & 0x01)) buffer[pos+xb+1] |= 0x80;
                  }

                  pos += newwidth;
                  gfx += oldwidth;
               }

               Cache->Chars[i].Width++;
               Cache->Chars[i].Advance++;
            }
         }

         *Data = buffer;
         fontdata = buffer;
      }
      else return(ERR_AllocMemory);
   }

   if (Flags & FTF_ITALIC) {
      DPrintF("Init:","Converting font graphic to italic.");

      size = 0;
      extra = Cache->Header.pixel_height>>2;

      for (i=0; i < 256; i++) {
         if (Cache->Chars[i].Width) size += Cache->Header.pixel_height * ((Cache->Chars[i].Width+7+extra)>>3);
      }

      if (AllocMemory(size, MEM_UNTRACKED, &buffer, NULL) IS ERR_Okay) {
         pos = 0;
         for (i=0; i < 256; i++) {
            if (Cache->Chars[i].Width) {
               gfx = fontdata + Cache->Chars[i].Offset;
               Cache->Chars[i].Offset = pos;

               oldwidth = (Cache->Chars[i].Width+7)>>3;
               newwidth = (Cache->Chars[i].Width+7+extra)>>3;
               italic = Cache->Header.pixel_height;
               dest = buffer + pos;
               for (y=0; y < Cache->Header.pixel_height; y++) {
                  dx = italic>>2;
                  for (sx=0; sx < Cache->Chars[i].Width; sx++) {
                     if (gfx[sx>>3] & (0x80>>(sx & 0x07))) {
                        dest[dx>>3] |= (0x80>>(dx & 0x07));
                     }
                     dx++;
                  }

                  pos  += newwidth;
                  dest += newwidth;
                  gfx  += oldwidth;
                  italic--;
               }

               Cache->Chars[i].Width += extra;
            }
         }

         if (*Data) FreeMemory(*Data);
         *Data = buffer;
      }
      else return(ERR_AllocMemory);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: GenerateFontOutline()
*/

static ERROR GenerateFontOutline(struct Font *Self, struct BitmapCache *Cache, UBYTE **Outline)
{
   UBYTE *buffer, *gfx, *dest;
   LONG size;
   UWORD i, pos;
   WORD oldwidth, newwidth, dx, sx, sy;

   DPrintF("Init:","Generating font outline.");

   size = 0;
   for (i=0; i < 256; i++) {
      if (Cache->Chars[i].Width) size += (Cache->Header.pixel_height+2) * ((Cache->Chars[i].Width+9)>>3);
   }

   if (AllocMemory(size, MEM_UNTRACKED, &buffer, NULL) IS ERR_Okay) {
      pos = 0;
      for (i=0; i < 256; i++) {
         if (Cache->Chars[i].Width) {
            gfx = Cache->Data + Cache->Chars[i].Offset;
            Cache->Chars[i].OutlineOffset = pos;

            oldwidth = (Cache->Chars[i].Width+7)>>3;
            newwidth = (Cache->Chars[i].Width+9)>>3;

            dest = buffer + pos;

            dest += newwidth; /* Start ahead of line 0 */
            for (sy=0; sy < Cache->Header.pixel_height; sy++) {
               dx = 1;
               for (sx=0; sx < Cache->Chars[i].Width; sx++) {
                  if (gfx[sx>>3] & (0x80>>(sx & 0x07))) {
                     if ((sx >= Cache->Chars[i].Width-1) OR (!(gfx[(sx+1)>>3] & (0x80>>((sx+1) & 0x07))))) dest[(dx+1)>>3] |= (0x80>>((dx+1) & 0x07));
                     if ((sx IS 0) OR (!(gfx[(sx-1)>>3] & (0x80>>((sx-1) & 0x07))))) dest[(dx-1)>>3] |= (0x80>>((dx-1) & 0x07));
                     if ((sy < 1) OR (!(gfx[(sx>>3)-oldwidth] & (0x80>>(sx & 0x07))))) dest[(dx>>3)-newwidth] |= (0x80>>(dx & 0x07));
                     if ((sy >= Cache->Header.pixel_height-1) OR (!(gfx[(sx>>3)+oldwidth] & (0x80>>(sx & 0x07))))) dest[(dx>>3)+newwidth] |= (0x80>>(dx & 0x07));
                  }
                  dx++;
               }

               pos  += newwidth;
               dest += newwidth;
               gfx  += oldwidth;
            }
            pos += newwidth * 2;
         }
      }

      *Outline = buffer;
      return(ERR_Okay);
   }
   else return(ERR_AllocMemory);
}

/*****************************************************************************
** Internal: CheckBitmapCache()
*/

static WORD CheckBitmapCache(struct Font *Self, LONG Style)
{
   WORD index;

   if (!glBitmapCache) return(-1);

   for (index=0; index < glCacheSize; index++) {
      if (glBitmapCache[index].Data) {
         if (StrCompare(glBitmapCache[index].Location, Self->Location, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            if (glBitmapCache[index].StyleFlags IS Style) {
               if (Self->Point IS glBitmapCache[index].Header.nominal_point_size) {
                  DPrintF("Init:","[Font] Exists in cache (count %d) %s : %s", glBitmapCache[index].OpenCount, glBitmapCache[index].Location, Self->prvStyle);
                  return(index);
               }
               //else DPrintF("@font:","Failed point check %d / %d", Self->Point, glBitmapCache[index].Header.nominal_point_size);
            }
            //else DPrintF("@font:","Failed style check $%.8x / $%.8x", Self->Flags, glBitmapCache[index].StyleFlags);
         }
      }
   }

   return(-1);
}

/*****************************************************************************
** Internal: CacheBitmapFont()
*/

static WORD CacheBitmapFont(struct Font *Self, OBJECTPTR file, LONG offset, struct winfnt_header_fields *face)
{
   struct acRead read;
   WORD index, i;
   LONG j, k, pos, bytewidth, size;

   DPrintF("Font:","Loading font into cache %s : %d : %s", Self->Location, face->nominal_point_size, Self->prvStyle);

   /*** Load the font into a new bitmap cache entry ***/

   if (!glBitmapCache) {
      if (AllocPrivateMemory(sizeof(struct BitmapCache) * 10, MEM_UNTRACKED, (void **)&glBitmapCache) IS ERR_Okay) {
         glCacheSize = 10;
         index = 0;
      }
      else return(-1);
   }
   else {
      for (index=0; index < glCacheSize; index++) {
         if ((!glBitmapCache[index].Data) AND (!glBitmapCache[index].OpenCount)) break;
      }

      /*** Expand the cache if there is not enough room ***/

      if (index >= glCacheSize) {
         if (ReallocMemory(glBitmapCache, sizeof(struct BitmapCache) * (glCacheSize + 10), &glBitmapCache, NULL) IS ERR_Okay) {
            index = glCacheSize;
            glCacheSize += 10;
         }
         else return(-1);
      }
   }

   CopyMemory(face, &glBitmapCache[index].Header, sizeof(struct winfnt_header_fields));

   /*** Record the style of the cached font ***/

   if (StrCompare("Bold", Self->prvStyle, 0, STR_MATCHLENGTH) IS ERR_Okay) glBitmapCache[index].StyleFlags = FTF_BOLD;
   else if (StrCompare("Italic", Self->prvStyle, 0, STR_MATCHLENGTH) IS ERR_Okay) glBitmapCache[index].StyleFlags = FTF_ITALIC;
   else if (StrCompare("Bold Italic", Self->prvStyle, 0, STR_MATCHLENGTH) IS ERR_Okay) glBitmapCache[index].StyleFlags = FTF_BOLD|FTF_ITALIC;
   else glBitmapCache[index].StyleFlags = NULL;

   StrCopy(Self->Location, glBitmapCache[index].Location, sizeof(glBitmapCache[index].Location));

   /*** Read character information from the font file ***/

   ActionTags(AC_Seek, file, (DOUBLE)(offset + 118), POS_START);

   if (face->version IS 0x300) {
      j = face->first_char;
      for (i=0; i < face->last_char - face->first_char + 1; i++) {
         glBitmapCache[index].Chars[j].Width   = ReadWordLE(file);
         glBitmapCache[index].Chars[j].Advance = glBitmapCache[index].Chars[j].Width;
         glBitmapCache[index].Chars[j].Offset  = ReadLongLE(file) - face->bits_offset; /* Long */
         j++;
      }
   }
   else {
      j = face->first_char;
      for (i=0; i < face->last_char - face->first_char + 1; i++) {
         glBitmapCache[index].Chars[j].Width   = ReadWordLE(file);
         glBitmapCache[index].Chars[j].Advance = glBitmapCache[index].Chars[j].Width;
         glBitmapCache[index].Chars[j].Offset  = ReadWordLE(file) - face->bits_offset; /* Word */
         j++;
      }
   }

   /*** Extract graphical data from the font file ***/

   size = face->file_size - face->bits_offset;

   if (AllocMemory(size, MEM_UNTRACKED, &glBitmapCache[index].Data, NULL) IS ERR_Okay) {
      ActionTags(AC_Seek, file, (DOUBLE)(offset + face->bits_offset), POS_START);

      read.Buffer = glBitmapCache[index].Data;
      read.Length = size;
      if ((Action(AC_Read, file, &read) IS ERR_Okay) AND (read.Result IS size)) {

         /* Convert the graphics format for wide characters from
         ** column-first format to row-first format.
         */

         for (i=0; i < 256; i++) {
            if (!glBitmapCache[index].Chars[i].Width) continue;

            j = ((glBitmapCache[index].Chars[i].Width+7)>>3) * face->pixel_height;

            if (glBitmapCache[index].Chars[i].Width > 8) {
               UBYTE *gfx, buffer[j];

               ClearMemory(buffer, j);

               gfx = glBitmapCache[index].Data + glBitmapCache[index].Chars[i].Offset;
               bytewidth = (glBitmapCache[index].Chars[i].Width + 7)>>3;
               pos = 0;
               for (k=0; k < face->pixel_height; k++) {
                  for (j=0; j < bytewidth; j++) {
                     buffer[pos++] = gfx[k + (j * face->pixel_height)];
                  }
               }

               CopyMemory(buffer, gfx, pos);
            }
         }

         return(index);
      }
      else {
         DPrintF("@Init:","[Font] Failed to read %d bytes (got %d).", size, read.Result);
         return(-1);
      }
   }
   else return(-1);
}
