/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Clock
Version:   1.0
ID:        ID_CLOCK
Status:    Stable
Category:  GUI
Date:      February 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 1999-2003.  All rights reserved.
Short:     This class creates a real-time clock/time display.
Keywords:  clock, time, timer, watch

###DESCRIPTION###
<p>The Clock class draws text-based information that always reflects the
current time.  It is typically used as a digital clock positioned on the
desktop so that the user has quick access to the current time.</p>

<p>There are no compulsory settings for creating a clock object, but in
most cases you will want to set the position and its graphical style, which
can be altered by accessing the <a href="#Font">Font</a> attributes.  You
should also set the <a href="#Format">Format</a> of the font if you want
something other than a 24 hour clock display.  Here is a DML based
example:</p>

<pre>
   &lt;clock format="d Mmm yyyy hh:nn" colour="#ffffff" x="100" y="5"/&gt;
</pre>

<p>Like all graphical objects, the clock should be contained by a drawable
area such as a Render object, or you can set the Drawable field to point
to the graphics area.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/misc/time.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_CLOCK  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_CLOCK, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 1999-2003.  All rights reserved.",
   "February 2003",
   "Clock"
};

static struct KernelBase  *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static OBJECTPTR ClockClass = 0;
static OBJECTPTR StringsMod = 0;
static OBJECTID TimeID = 0;

/*****************************************************************************
** Clock class definition.
*/

#define LEN_STRING 100
#define LEN_FORMAT 40

struct Clock {
   OBJECT_HEADER
   OBJECTID DrawableID;
   LONG     AlignFlags;
   LONG     Flags;
   struct Font *Font;

   /*** Private field store ***/
   BYTE  Format[LEN_FORMAT];
   BYTE  String[LEN_STRING];
   LONG  Dimensions;
   FLOAT XCoord;
   FLOAT YCoord;
   FLOAT XOffset;
   FLOAT YOffset;
   FLOAT Height;
   FLOAT Width;
};

#define CLF_VISIBLE 0x00000001

static ERROR GET_Format(struct Clock *, STRING *);
static ERROR GET_Height(struct Clock *, struct Variable *);
static ERROR GET_Width(struct Clock *, struct Variable *);
static ERROR GET_X(struct Clock *, struct Variable *);
static ERROR GET_Y(struct Clock *, struct Variable *);
static ERROR GET_XOffset(struct Clock *, struct Variable *);
static ERROR GET_YOffset(struct Clock *, struct Variable *);

static ERROR SET_Align(struct Clock *, LONG *);
static ERROR SET_Format(struct Clock *, STRING);
static ERROR SET_Height(struct Clock *, struct Variable *);
static ERROR SET_Width(struct Clock *, struct Variable *);
static ERROR SET_X(struct Clock *, struct Variable *);
static ERROR SET_XOffset(struct Clock *, struct Variable *);
static ERROR SET_Y(struct Clock *, struct Variable *);
static ERROR SET_YOffset(struct Clock *, struct Variable *);

static struct FieldDef AlignFlags[] = {
   { "RIGHT",      ALIGN_RIGHT   },    { "LEFT",     ALIGN_LEFT     },
   { "BOTTOM",     ALIGN_BOTTOM  },    { "TOP",      ALIGN_TOP      },
   { "CENTER",     ALIGN_CENTER  },    { "MIDDLE",   ALIGN_MIDDLE   },
   { "HORIZONTAL", ALIGN_HORIZONTAL }, { "VERTICAL", ALIGN_VERTICAL },
   { NULL, NULL }
};

static struct FieldDef ClockFlags[] = {
   { "VISIBLE", CLF_VISIBLE },
   { NULL, NULL }
};

static struct FieldArray ClockFields[] = {
   { "Drawable", 0, FDF_OBJECTID|FDF_RW,  ID_RENDER,         NULL, NULL },
   { "Align",    0, FDF_LONGFLAGS|FDF_RW, (LONG)&AlignFlags, NULL, SET_Align },
   { "Flags",    0, FDF_LONGFLAGS|FDF_RW, (LONG)&ClockFlags, NULL, NULL },
   { "Font",     0, FDF_CHILD|FDF_R,      ID_FONT,           NULL, NULL },

   /*** Virtual fields ***/
   { "Format",   0, FDF_STRING|FDF_RW, 0, GET_Format, SET_Format },

   /*** Variable fields ***/
   { "Height",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",   0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X, SET_X },
   { "XCoord",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X, SET_X },
   { "XOffset", 0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_W,  0, GET_XOffset, SET_XOffset },
   { "Y",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y, SET_Y },
   { "YCoord",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y, SET_Y },
   { "YOffset", 0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_W,  0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*****************************************************************************
** Clock actions.
*/

static ERROR CLOCK_Free(struct Clock *, APTR);
static ERROR CLOCK_NewObject(struct Clock *, APTR);
static ERROR CLOCK_Hide(struct Clock *, APTR);
static ERROR CLOCK_Init(struct Clock *, APTR);
static ERROR CLOCK_Show(struct Clock *, APTR);
static ERROR CLOCK_Timer(struct Clock *, struct acTimer *);

static struct ActionArray ClockActions[] = {
   { AC_Free,      CLOCK_Free },
   { AC_Hide,      CLOCK_Hide },
   { AC_Init,      CLOCK_Init },
   { AC_NewObject, CLOCK_NewObject },
   { AC_Show,      CLOCK_Show },
   { AC_Timer,     CLOCK_Timer },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL, FID_Name|TSTRING, "strings", TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (NewPublicObject(ID_TIME, NULL, &TimeID) IS ERR_Okay) {
      ActionMsg(AC_Init, TimeID, NULL);
   }
   else return(SendError(ERH_Module, ERR_NewObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ClockClass, NULL,
      FID_BaseClassID|TLONG, ID_CLOCK,
      FID_Version|TFLOAT,    VER_CLOCK,
      FID_Name|TSTRING,      "Clock",
      FID_Category|TLONG,    CCF_GUI,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN|CLF_PRIVATEONLY,
      FID_Actions|TPTR,      ClockActions,
      FID_Fields|TPTR,       ClockFields,
      FID_Size|TLONG,        sizeof(struct Clock),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (TimeID)     { ActionMsg(AC_Free, TimeID, NULL);  TimeID     = NULL; }
   if (StringsMod) { Action(AC_Free, StringsMod, NULL); StringsMod = NULL; }
   if (ClockClass) { Action(AC_Free, ClockClass, NULL); ClockClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: DrawClock()
*/

static void DrawClock(struct Clock *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct Font *Font;
   struct Variable var;
   LONG strwidth, clipleft, clipright, cliptop, clipbottom, x, y, width, height;

   if (!(Self->Flags & CLF_VISIBLE)) return;

   if (!(Font = Self->Font)) return;

   Font->Bitmap = Bitmap;

   var.Type = FD_DOUBLE;
   GET_X(Self, &var); x = var.Double;
   GET_Y(Self, &var); y = var.Double;
   GET_Width(Self, &var);  width = var.Double;
   GET_Height(Self, &var); height = var.Double;

   clipleft   = Bitmap->ClipLeft;
   clipright  = Bitmap->ClipRight;
   cliptop    = Bitmap->ClipTop;
   clipbottom = Bitmap->ClipBottom;
   if (x > Bitmap->ClipLeft) Bitmap->ClipLeft = x;
   if (y > Bitmap->ClipTop)  Bitmap->ClipTop  = y;
   if (x + width < Bitmap->ClipRight) Bitmap->ClipRight = x + width;
   if (y + height < Bitmap->ClipBottom) Bitmap->ClipBottom = y + height;

   SetField(Font, FID_String, FT_STRING, Self->String);

   /*** Calculate the font alignment within the clock's drawable ***/

   GetField(Font, FID_Width, FT_LONG, &strwidth);

   Font->XCoord = x;
   Font->YCoord = y;
   if (Self->AlignFlags & ALIGN_RIGHT)      Font->XCoord = x + width - strwidth;
   if (Self->AlignFlags & ALIGN_HORIZONTAL) Font->XCoord = x + ((width - strwidth) / 2);
   if (Self->AlignFlags & ALIGN_BOTTOM)     Font->YCoord = y + height - Font->MaxHeight;
   if (Self->AlignFlags & ALIGN_VERTICAL)   Font->YCoord = y + ((height - Font->MaxHeight) / 2);

   Action(AC_Draw, Font, NULL);

   Bitmap->ClipLeft   = clipleft;
   Bitmap->ClipRight  = clipright;
   Bitmap->ClipTop    = cliptop;
   Bitmap->ClipBottom = clipbottom;
}

/*****************************************************************************
** Clock: Free
*/

static ERROR CLOCK_Free(struct Clock *Self, APTR Void)
{
   if (Self->DrawableID) {
      OBJECTPTR render;
      if (AccessObject(Self->DrawableID, 3000, &render) IS ERR_Okay) {
         ActionTags(MT_SetRenderCallback, render, Self, NULL);
         ReleaseObject(render);
      }
   }

   if (Self->Font) {
      Action(AC_Free, Self->Font, NULL);
      Self->Font = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Hides a clock display.
###END###

*****************************************************************************/

static ERROR CLOCK_Hide(struct Clock *Self, APTR Void)
{
   if (Self->Flags & CLF_VISIBLE) {
      Self->Flags &= ~CLF_VISIBLE;
      return(ActionMsg(AC_Draw, Self->DrawableID, NULL));
   }
   else return(ERR_Okay);
}

/*****************************************************************************
** Clock: Init
*/

static ERROR CLOCK_Init(struct Clock *Self, APTR Void)
{
   struct Variable value;
   OBJECTPTR render;

   /*** Find the Render object that we are associated with ***/

   if (!Self->DrawableID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = OwnerID;
   }

   /*** If coordinates are missing, set them to zero ***/

   value.Type = FD_DOUBLE;
   value.Double = 0;
   if ((!(Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)))) {
      if (!((Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)) AND (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)))) {
         SET_X(Self, &value);
      }
   }

   if ((!(Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)))) {
      if (!((Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)) AND (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)))) {
         SET_Y(Self, &value);
      }
   }

   if ((!(Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) AND (!(Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)))) {
      SET_XOffset(Self, &value);
   }

   if ((!(Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) AND (!(Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)))) {
      SET_YOffset(Self, &value);
   }

   /*** Initialise the Font ***/

   Self->Font->Flags |= FTF_DIRECTXY;
   if (Action(AC_Init, Self->Font, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Init));
   }

   if (AccessObject(Self->DrawableID, 5000, &render) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, render, Self, &DrawClock);
      ReleaseObject(render);
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   SubscribeTimer(Self->Head.UniqueID, 1000);

   CLOCK_Timer(Self, NULL); /* Call the timer function to build the clock string */
   return(ERR_Okay);
}

/*****************************************************************************
** Clock: NewObject
*/

static ERROR CLOCK_NewObject(struct Clock *Self, APTR Void)
{
   if (NewObject(ID_FONT, NF_CHILD, &Self->Font, NULL) IS ERR_Okay) {
      SetFields(Self->Font, FID_Face|TSTRING, "sans serif", TAGEND);
      Self->Flags |= CLF_VISIBLE;

      StrCopy("hh:nn", Self->Format, COPY_ALL);
      return(ERR_Okay);
   }
   else return(ERR_NewObject);
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Shows a clock display.
###END###

*****************************************************************************/

static ERROR CLOCK_Show(struct Clock *Self, APTR Void)
{
   if (!(Self->Flags & CLF_VISIBLE)) {
      Self->Flags |= CLF_VISIBLE;
      return(ActionMsg(AC_Draw, Self->DrawableID, NULL));
   }
   else return(ERR_Okay);
}

/*****************************************************************************
** Clock: Timer
*/

static ERROR CLOCK_Timer(struct Clock *Self, struct acTimer *Args)
{
   struct Time *Time;
   UBYTE buffer[LEN_STRING];

   if (Self->Flags & CLF_VISIBLE) {

      if (AccessObject(TimeID, 0, (OBJECTPTR *)&Time) IS ERR_Okay) {
         Action(AC_Query, Time, NULL);

         StrCopy(Self->String, buffer, sizeof(buffer));

         StrFormatDate(Self->String, sizeof(Self->String), Self->Format, Time);

         if (StrCompare(buffer, Self->String, 0, STR_MATCHCASE|STR_MATCHLENGTH) != ERR_Okay) {
            DelayMsg(AC_Draw, Self->DrawableID, NULL);
         }

         ReleaseObject(Time);
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Align
Short:  This field allows you to align a clock's text within its container.
Type:   LONG/FLAGS
Prefix: ALIGN_
Status: Write

###DESCRIPTION###
<p>If you would like to set an abstract position for the Clock, you can
give it an alignment.  Note that setting the alignment overrides any
settings in related coordinate fields.  Valid alignment flags are as
follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">BOTTOM</td><td>Aligns the Clock to the bottom of the drawable area.</td></tr>
<tr><td width="10%">CENTER</td><td>Centers the Clock on both the vertical and horizontal axis.</td></tr>
<tr><td width="10%">LEFT</td><td>Aligns the Clock to the left (horizontal coordinate 0).</td></tr>
<tr><td width="10%">MIDDLE</td><td>Synonym for the CENTER flag.</td></tr>
<tr><td width="10%">RIGHT</td><td>Aligns the Clock so that the last character is positioned flush against the right.</td></tr>
<tr><td width="10%">TOP</td><td>Aligns the Clock to the top (vertical coordinate 0).</td></tr>
<tr><td width="10%">HORIZONTAL</td><td>Centers the Clock on the horizontal axis.</td></tr>
<tr><td width="10%">VERTICAL</td><td>Centers the Clock on the vertical axis.</td></tr>
</table>

###END###

*****************************************************************************/

static ERROR SET_Align(struct Clock *Self, LONG *Alignment)
{
   Self->AlignFlags = *Alignment;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  This field can be set to define the drawable area for Clock graphics.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>In most cases, when you create a new Clock it should be contained by a
drawable area, such as a Render object.  If you do this correctly then
the Clock will detect the drawable area and set the Drawable field to the
correct object. However, there may be occasions where you want to
initialise a Clock object to a non-graphical container.  If this is the
case then you must set the Drawable field to point to a valid drawable
area, such as a Render object.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Font
Short:  Points to a Font that controls the drawing of text for a clock object.
Type:   struct Font *
Status: Read

###DESCRIPTION###
<p>To set the face, colour and other attributes of a clock's graphics,
you need to read the Font field and write your settings to the Font object
prior to initialisation.  For a list of all the fields that can be set, please
refer to the Font class manual.</p>

<p>You are strongly discouraged from writing information to a Font object
after the initialistion of a clock, although you may read information from
the Font after initialisation.</p>

###SEE ALSO###
Class: Font

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Format
Short:  The display format for a clock's time can be defined in this field.
Type:   STRING
Status: Read/Write

###DESCRIPTION###
<p>Set the Format field to define how the time information is printed to the
display.  Time formatting follows typical convention, as illustrated in the
following format table:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">String</td><td class="tableheader" bgcolor="#eeeeee">Example</td></tr>
<tr><td width="10%">h</td><td>24-Hour: 9</td></tr>
<tr><td width="10%">hh</td><td>24-Hour: 09</td></tr>
<tr><td width="10%">H</td><td>12-Hour: 9</td></tr>
<tr><td width="10%">HH</td><td>12-Hour: 09</td></tr>
<tr><td width="10%">n</td><td>Minute: 7</td></tr>
<tr><td width="10%">nn</td><td>Minute: 07</td></tr>
<tr><td width="10%">s</td><td>Second: 42</td></tr>
<tr><td width="10%">ss</td><td>Second: 42</td></tr>
<tr><td width="10%">d</td><td>Day: 3</td></tr>
<tr><td width="10%">dd</td><td>Day: 03</td></tr>
<tr><td width="10%">D</td><td>Day of week: T</td></tr>
<tr><td width="10%">ddd</td><td>Day of week: sat</td></tr>
<tr><td width="10%">dddd</td><td>Day of week: saturday</td></tr>
<tr><td width="10%">m</td><td>Month: 11</td></tr>
<tr><td width="10%">mm</td><td>Month: 11</td></tr>
<tr><td width="10%">mmm</td><td>Month: nov</td></tr>
<tr><td width="10%">mmmm</td><td>Month: november</td></tr>
<tr><td width="10%">yy</td><td>Year: 03</td></tr>
<tr><td width="10%">yyyy</td><td>Year: 2003</td></tr>
</table>

<p>By default, strings are printed in lower-case.  For example, "ddd" prints
"sat".  If you want to capitalise the characters, use capitals in the string
formatting instead of the default lower case.  For example "Ddd" prints "Sat"
and "DDD" prints "SAT".</p>

<p>Where the interpreter encounters unrecognised characters, it will print
them out unaltered.  For example "hh:nn:ss" might print "05:34:19" with the
colons intact.  If you need to include characters such as 'h' or 'd' and
want to prevent them from being translated, preceed them with the backslash
character.  For instance to print "0800 hours", we would use the format
"hh00 \hours".</p>

###END###

*****************************************************************************/

static ERROR GET_Format(struct Clock *Self, STRING *Format)
{
   *Format = Self->Format;
   return(ERR_Okay);
}

static ERROR SET_Format(struct Clock *Self, STRING Format)
{
   StrCopy(Format, Self->Format, sizeof(Self->Format));
   return(ERR_Okay);
}


/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of a clock object's drawable region.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A clock's drawing region can be given a fixed or relative height by
setting this field to the desired value.  To set a relative height, use the
FD_PERCENT flag when setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Height(struct Clock *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerheight, ycoord, value;

   if (!(Self->Dimensions & DMF_VERTICAL_FLAGS)) {
      value = Self->Font->LineSpacing * Self->Font->LineCount;
   }
   else if (Self->Dimensions & DMF_FIXED_HEIGHT) value = Self->Height;
   else if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ObjectError(Self, ERH_GetField, ERR_GetField));

      if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         value = Self->Height * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_Y) ycoord = Self->YCoord;
         else ycoord = containerheight * Self->YCoord / 100;

         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - ycoord - Self->YOffset;
         else value = containerheight - ycoord - (containerheight * Self->YOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Height(struct Clock *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Clock] A height of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Height = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[Clock] A height of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Height = Value->Large;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a clock's drawable region.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A clock's drawing region can be given a fixed or relative width by
setting this field to the desired value.  To set a relative width, use the
FD_PERCENT flag when setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Width(struct Clock *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, xcoord, value;

   if (!(Self->Dimensions & DMF_HORIZONTAL_FLAGS)) {
      SetField(Self->Font, FID_String, FT_STRING, Self->String);
      if (GetField(Self->Font, FID_Width, FT_DOUBLE, &value) != ERR_Okay) return(ERR_Failed);
   }
   else if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ObjectError(Self, ERH_GetField, ERR_GetField));

      if (Self->Dimensions & DMF_FIXED_WIDTH) value = Self->Width;
      else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         value = Self->Width * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_X) xcoord = Self->XCoord;
         else xcoord = containerwidth * Self->XCoord / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - xcoord - Self->XOffset;
         else value = containerwidth - xcoord - (containerwidth * Self->XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Width(struct Clock *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Clock] A width of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Width = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[Clock] A width of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Width = Value->Large;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  Defines the horizontal position of a clock's graphic.
Type:   DOUBLE/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>This field determines the horizontal position of a clock's graphic from
the left of its container.  The value can be set as a fixed pixel coordinate,
or as a percentage if the FD_PERCENT field is used.</p>

###SEE ALSO###
Fields: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Clock *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (Self->Dimensions & DMF_FIXED_X) value = Self->XCoord;
   else if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_RELATIVE_X) {
         value = Self->XCoord * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - width - Self->XOffset;
         else value = containerwidth - width - (containerwidth * Self->XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct Clock *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->XCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  Defines the horziontal offset of a clock's graphic.
Type:   DOUBLE/PERCENTAGE
Status: Write

###DESCRIPTION###
<p>You can have a clock's graphic offset against the right of its container
by setting this field.  A setting of 0 would place the clock at the right
edge, while a setting of 10 would place it 10 pixels off from the right
edge. Negative values can move the clock outside of the container's
boundaries.</p>

<p>This field may also be set as a percentage if the FD_PERCENT flag is used.
Use of a percentage will cause the position of the clock to be calculated
relative to the width of the container.</p>

<p>Writing to this field will always turn off any previous setting in the
<a href="#XCoord">XCoord</a> field.</p>

###SEE ALSO###
Fields: XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Clock *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_XOFFSET) value = Self->XOffset;
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         value = Self->XOffset * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_X) value = containerwidth - (Self->XCoord + width);
         else value = containerwidth - ((Self->XCoord * containerwidth / 100) + width);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Clock *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->XOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  Defines the vertical position of a clock's graphic.
Type:   DOUBLE/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>This field determines the vertical position of a clock's graphic from the
top of its container.  The value can be set as a fixed pixel coordinate, or
as a percentage if the FD_PERCENT field is used.</p>

###SEE ALSO###
Fields: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct Clock *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   if (Self->Dimensions & DMF_FIXED_Y) value = Self->YCoord;
   else if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_RELATIVE_Y) {
         value = Self->YCoord * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - height - Self->YOffset;
         else value = containerheight - height - (containerheight * Self->YOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct Clock *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->YCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  Defines the vertical offset of a clock's graphic.
Type:   DOUBLE/PERCENTAGE
Status: Write

###DESCRIPTION###
<p>You can have a clock's graphic offset against the bottom of its container
by setting this field.  A setting of 0 would place the clock at the bottom
edge, while a setting of 10 would place it 10 pixels off from the bottom
edge. Negative values can move the clock outside of the container's
boundaries.</p>

<p>This field may also be set as a percentage if the FD_PERCENT flag is used.
Use of a percentage will cause the position of the clock to be calculated
relative to the height of the container.</p>

<p>Writing to this field will always turn off any previous setting in the
<a href="#YCoord">YCoord</a> field.</p>

###SEE ALSO###
Fields: YCoord, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Clock *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   value = 0;
   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_YOFFSET) value = Self->YOffset;
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         value = Self->YOffset * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_Y) value = containerheight - (Self->YCoord + height);
         else value = containerheight - ((Self->YCoord * containerheight / 100) + height);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Clock *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->YOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
   return(ERR_Okay);
}
